from typing import Tuple

import numpy as np
import tensorflow as tf

from tfimm.layers import act_layer_factory, norm_layer_factory


class PromptEncoder(tf.keras.Model):
    def __init__(
        self,
        embed_dim: int,
        mask_hidden_dim: int,
        act_layer: str = "gelu",
        **kwargs,
    ) -> None:
        """
        Encodes prompts for input to SAM's mask decoder.

        Arguments:
            embed_dim: The prompts' embedding dimension.
            mask_hidden_dim: The number of hidden channels for encoding input masks.
            act_layer: Activation layer.
        """
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.mask_hidden_dim = mask_hidden_dim
        self.act_layer = act_layer

        self.pe_layer = PositionalEmbeddingRandom(
            embed_dim=self.embed_dim, name="pe_layer"
        )
        self.point_embeddings = None
        self.not_a_point_embed = None

        self.mask_downscaling = MaskDownscaling(
            embed_dim=self.embed_dim,
            mask_hidden_dim=self.mask_hidden_dim,
            act_layer=self.act_layer,
            name="mask_downscaling",
        )
        self.no_mask_embed = None

    @property
    def dummy_inputs(self):
        return {
            "points": tf.zeros((1, 1, 2), dtype=tf.float32),
            "labels": tf.zeros((1, 1), dtype=tf.int32),
            "boxes": tf.zeros((1, 1, 4), dtype=tf.float32),
            "masks": tf.zeros((1, 8, 8, 1), dtype=tf.float32),
        }

    def build(self, input_shape):
        nb_point_embeddings = 4  # foreground/background point + 2 box corners
        self.point_embeddings = [
            self.add_weight(
                shape=(1, self.embed_dim),
                initializer=tf.keras.initializers.RandomNormal(),
                trainable=True,
                name=f"point_embeddings/{j}/weight",
            )
            for j in range(nb_point_embeddings)
        ]
        self.not_a_point_embed = self.add_weight(
            shape=(1, self.embed_dim),
            initializer=tf.keras.initializers.RandomNormal(),
            trainable=True,
            name="not_a_point_embed/weight",
        )
        self.no_mask_embed = self.add_weight(
            shape=(1, self.embed_dim),
            initializer=tf.keras.initializers.RandomNormal(),
            trainable=True,
            name="no_mask_embed/weight",
        )

    def _embed_points(self, points: tf.Tensor, labels: tf.Tensor, input_size):
        """Embed point prompts."""
        points = points + 0.5  # Shift to center of pixel
        labels = tf.expand_dims(labels, axis=-1)  # (N, 1)
        embeddings = self.pe_layer.embed_points(points, input_size)
        embeddings += tf.where(
            labels == 0, self.point_embeddings[0], self.point_embeddings[1]
        )
        return embeddings

    def _embed_boxes(self, boxes: tf.Tensor, input_size):
        """Embed box promts."""
        # Shape of boxes is (N, M, 4)
        n, m, _ = tf.unstack(tf.shape(boxes))
        boxes = boxes + 0.5  # Shift to center of pixel
        corners = tf.reshape(boxes, (n * m, 2, 2))  # (N*M, 2, 2)
        embeddings = self.pe_layer.embed_points(
            corners, input_size  # (N*M, 2, embed_dim)
        )
        embeddings += tf.stack(
            [self.point_embeddings[2], self.point_embeddings[3]], axis=1
        )
        embeddings = tf.reshape(
            embeddings, (n, m, 2, self.embed_dim)
        )  # (N, M, 2, embed_dim)
        embeddings = tf.reshape(
            embeddings, (n, 2 * m, self.embed_dim)
        )  # (N, 2*M, embed_dim)
        return embeddings

    def _embed_masks(self, masks, training: bool = False):
        n, _, h, w = tf.unstack(tf.shape(masks))  # (N, M3, H', W')
        # If we have no masks, we return `no_mask_embed` broadcast across batch and
        # spatial dimensions.
        embeddings = tf.cond(
            tf.shape(masks)[1] == 0,
            lambda: tf.broadcast_to(
                input=tf.reshape(self.no_mask_embed, (1, 1, 1, -1)),
                shape=(n, h // 4, w // 4, self.embed_dim),
            ),
            lambda: self.mask_downscaling(masks, training=training),
        )
        return embeddings  # (N, H'', W'', D)

    def call(self, inputs, training=False):
        """
        Embeds different types of prompts, returning both sparse and dense embeddings.

        Arguments:
            inputs: Dictionary with the following entries.
                points: An (N, M1, 2) tensor of point coordinates.
                labels: An (N, M1) tensor of point labels. 1 indicates a foreground
                    point and 0 indicates a background point.
                boxes: An (N, M2, 4) tensor of bounding box coordinates.
                masks: An (N, M3, H', W') tensor for masks, where M3 can be 0 (no mask
                    to embed, tensor is empty). The spatial input size (H, W) should be
                    self.mask_size.
            training: Training or inference phase?

        Returns:
            Sparse embeddings of shape (N, M, D), where M is determined by the number
                of input points and boxes:
                 - M = 2*M2, if M1 = 0 (no points)
                 - M = M1 + 1, if M1 > 0 and M2 = 0 (points, no boxes) and
                 - M = M1 + 2*M2, if M1 > 0 and M2 > 0 (points and boxes)
            Dense mask embeddings of shape (N, H'', W'', D), where (H'', W'') is given
                by grid_size and D is embed_dim.
        """
        points = inputs["points"]
        labels = inputs["labels"]
        boxes = inputs["boxes"]
        masks = inputs["masks"]

        _, _, h, w = tf.unstack(tf.shape(masks))  # (N, M3, H', W')
        input_size = (4 * h, 4 * w)  # (H, W)

        point_embeddings = self._embed_points(points, labels, input_size)
        box_embeddings = self._embed_boxes(boxes, input_size)

        # If we have points, but no boxes, we append one `not_a_point_embed` embedding.
        # This is done in PyTorch via `pad=True` in `_embed_points()`.
        n = tf.shape(points)[0]
        pad_token = tf.cond(
            (tf.shape(points)[1] > 0) & (tf.shape(boxes)[1] == 0),
            lambda: tf.tile(  # We need this to get the right batch size
                input=tf.expand_dims(self.not_a_point_embed, axis=0),
                multiples=(n, 1, 1),
            ),
            lambda: tf.zeros((n, 0, self.embed_dim)),
        )

        sparse_embeddings = tf.concat(
            [point_embeddings, pad_token, box_embeddings], axis=1
        )
        dense_embeddings = self._embed_masks(masks, training=training)

        return sparse_embeddings, dense_embeddings

    def get_dense_pe(self, grid_size) -> tf.Tensor:
        """
        Returns the positional encoding used to encode point prompts, applied to a
        dense set of points the shape of the image encoding.

        Returns:
          Positional encoding with shape (*grid_size, embed_dim).
        """
        image_pe = self.pe_layer.embed_grid(grid_size)
        return image_pe


class MaskDownscaling(tf.keras.Model):
    def __init__(
        self,
        embed_dim: int,
        mask_hidden_dim: int,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.mask_hidden_dim = mask_hidden_dim
        self.act_layer = act_layer

        norm_layer = norm_layer_factory("layer_norm_eps_1e-6")
        act_layer = act_layer_factory(self.act_layer)

        self.conv1 = tf.keras.layers.Conv2D(
            filters=self.mask_hidden_dim // 4, kernel_size=2, strides=2, name="0"
        )
        self.norm1 = norm_layer(name="1")
        self.act1 = act_layer(name="2")
        self.conv2 = tf.keras.layers.Conv2D(
            filters=self.mask_hidden_dim, kernel_size=2, strides=2, name="3"
        )
        self.norm2 = norm_layer(name="4")
        self.act2 = act_layer(name="5")
        self.conv3 = tf.keras.layers.Conv2D(
            filters=self.embed_dim, kernel_size=1, name="6"
        )

    def call(self, inputs, training=False):
        x = inputs  # (N, M3, H', W')

        n, m, h, w = tf.unstack(tf.shape(x))
        x = tf.reshape(x, (n * m, h, w))  # (N*M3, H', W')
        x = tf.expand_dims(x, axis=-1)  # (N*M3, H', W', 1)

        x = self.conv1(x)
        x = self.norm1(x, training=training)
        x = self.act1(x)
        x = self.conv2(x)
        x = self.norm2(x, training=training)
        x = self.act2(x)
        x = self.conv3(x)  # (N*M3, H'', W'', D)

        _, h, w, d = tf.unstack(tf.shape(x))
        x = tf.reshape(x, (n, m, h, w, d))  # (N, M3, H'', W'', D)
        # If we are given multiple input masks, we sum the corresponding embeddings.
        x = tf.reduce_sum(x, axis=1)  # (N, H'', W'', D)

        return x


class PositionalEmbeddingRandom(tf.keras.layers.Layer):
    """
    Positional embedding using random spatial frequencies.
    """

    def __init__(
        self,
        embed_dim: int,
        scale: float = 1.0,
        **kwargs,
    ) -> None:
        """
        Args:
            embed_dim: Size of positional embeddings. Note that compared to the PyTorch
                code, embed_dim = 2 * num_pos_feats, but since we concatenate the
                x and y positional embeddings, the result has size embed_dim.
            scale: Norm of embeddings
        """
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.scale = scale

        self.positional_encoding_gaussian_matrix = None

    def build(self, input_shape):
        self.positional_encoding_gaussian_matrix = self.add_weight(
            shape=(2, self.embed_dim // 2),
            initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=self.scale),
            trainable=False,
            name="positional_encoding_gaussian_matrix",
        )

    def call(self, x):
        """
        Positionally embed points that are normalised to [0, 1].

        Args:
            x: Tensor of shape (X, 2), X can be arbitrary.

        Returns:
            Tensor of shape (X, embed_dim).
        """
        x = 2 * x - 1  # (X, 2)
        x = tf.matmul(x, self.positional_encoding_gaussian_matrix)  # (X, D // 2)
        x = 2 * np.pi * x
        embed = tf.concat([tf.sin(x), tf.cos(x)], axis=-1)  # (X, D)
        return embed

    def embed_grid(self, size: Tuple[int, int]):
        """Generate positional embedding for a grid of the specified size."""
        h, w = size
        grid = tf.ones((h, w), dtype=tf.float32)
        x = (tf.cumsum(grid, axis=1) - 0.5) / tf.cast(w, tf.float32)
        y = (tf.cumsum(grid, axis=0) - 0.5) / tf.cast(h, tf.float32)
        points = tf.stack([x, y], axis=-1)
        return self(points)

    def embed_points(self, points, image_size):
        """Positionally embed points that are not normalized to [0,1]."""
        x = points[..., 0] / tf.cast(image_size[1], tf.float32)
        y = points[..., 1] / tf.cast(image_size[0], tf.float32)
        points = tf.stack([x, y], axis=-1)
        return self(points)



"""
This module includes pytorch architectures that are not part of timm, but use the timm
type registry system. We include them here so we can use them in unit tests.
"""
from .poolformer import *  # noqa: F401
from .pvt import *  # noqa: F401
from .pvt_v2 import *  # noqa: F401



"""
TensorFlow implementation of ResNets

Based on timm/models/resnet.py by Ross Wightman.

It includes the following models:
- Resnets from the PyTorch model hub
- ResNets trained by Ross Wightman
- Models pretrained on weakly-supervised data, finetuned on ImageNet
  Paper: Exploring the Limits of Weakly Supervised Pretraining
  Link: https://arxiv.org/abs/1805.00932
- Models pretrained in semi-supervised way on YFCC100M, finetuned on ImageNet
  Paper: Billion-scale Semi-Supervised Learning for Image Classification
  Link: https://arxiv.org/abs/1905.00546
- ResNets with ECA layers
  Paper: ECA-Net: Efficient Channel Attention for Deep Convolutional Neural Networks
  Link: https://arxiv.org/pdf/1910.03151.pdf
- ResNets with anti-aliasing layers
  Paper: Making Convolutional Networks Shift-Invariant Again
  Link: https://arxiv.org/pdf/1904.11486.pdf
- Models from the "Revisiting ResNets" paper
  Paper: Revisiting ResNets
  Link: https://arxiv.org/abs/2103.07579
- ResNets with a squeeze-and-excitation layer
  Paper: Squeeze-and-Excitation Networks
  Link: https://arxiv.org/abs/1709.01507

Copyright 2021 Martins Bruveris
Copyright 2021 Ross Wightman
"""

import math
from dataclasses import dataclass
from typing import List, Optional, Tuple

import tensorflow as tf

from tfimm.layers import (
    BlurPool2D,
    ClassifierHead,
    DropPath,
    act_layer_factory,
    attn_layer_factory,
    norm_layer_factory,
)
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# model_registry will add each entrypoint fn to this
__all__ = ["ResNet", "ResNetConfig", "BasicBlock"]


# TODO: Implement DropBlock, drop_block_rate in timm
#       See: https://arxiv.org/pdf/1810.12890.pdf)
@dataclass
class ResNetConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    # Residual blocks
    block: str = "basic_block"
    nb_blocks: Tuple = (2, 2, 2, 2)
    nb_channels: Tuple = (64, 128, 256, 512)
    cardinality: int = 1  # Number of groups in bottleneck conv
    base_width: int = 64  # Determines number of channels in block
    downsample_mode: str = "conv"
    zero_init_last_bn: bool = True
    # Stem
    stem_width: int = 64
    stem_type: str = ""
    replace_stem_pool: bool = False
    # Other params
    block_reduce_first: int = 1
    down_kernel_size: int = 1
    act_layer: str = "relu"
    norm_layer: str = "batch_norm"
    aa_layer: str = ""
    attn_layer: str = ""
    se_ratio: float = 0.0625
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Head
    global_pool: str = "avg"
    # Parameters for inference
    test_input_size: Optional[Tuple[int, int]] = None
    pool_size: int = 7  # For test-time pooling (not implemented yet)
    crop_pct: float = 0.875
    interpolation: str = "bilinear"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "conv1"
    classifier: str = "fc"

    def __post_init__(self):
        if self.test_input_size is None:
            self.test_input_size = self.input_size


class BasicBlock(tf.keras.layers.Layer):
    expansion = 1

    def __init__(
        self,
        cfg: ResNetConfig,
        nb_channels: int,
        stride: int,
        drop_path_rate: float,
        downsample_layer,
        **kwargs,
    ):
        super().__init__(**kwargs)
        assert cfg.cardinality == 1, "BasicBlock only supports cardinality of 1"
        assert cfg.base_width == 64, "BasicBlock does not support changing base width"

        self.cfg = cfg
        self.downsample_layer = downsample_layer
        self.act_layer = act_layer_factory(cfg.act_layer)
        self.norm_layer = norm_layer_factory(cfg.norm_layer)
        self.attn_layer = attn_layer_factory(cfg.attn_layer)

        # Num channels after first conv
        first_planes = nb_channels // cfg.block_reduce_first
        out_planes = nb_channels * self.expansion  # Num channels after second conv
        use_aa = cfg.aa_layer and stride == 2

        self.pad1 = tf.keras.layers.ZeroPadding2D(padding=1)
        self.conv1 = tf.keras.layers.Conv2D(
            filters=first_planes,
            kernel_size=3,
            # If we use anti-aliasing, the anti-aliasing layer takes care of strides
            strides=1 if use_aa else stride,
            use_bias=False,
            name="conv1",
        )
        self.bn1 = self.norm_layer(name="bn1")
        self.act1 = self.act_layer()
        self.aa = BlurPool2D(stride=stride) if use_aa else None

        self.pad2 = tf.keras.layers.ZeroPadding2D(padding=1)
        self.conv2 = tf.keras.layers.Conv2D(
            filters=out_planes,
            kernel_size=3,
            use_bias=False,
            name="conv2",
        )
        initializer = "zeros" if cfg.zero_init_last_bn else "ones"
        if cfg.norm_layer == "batch_norm":
            # Only batch norm layer has moving_variance_initializer parameter
            self.bn2 = self.norm_layer(
                gamma_initializer=initializer,
                moving_variance_initializer=initializer,
                name="bn2",
            )
        else:
            self.bn2 = self.norm_layer(gamma_initializer=initializer, name="bn2")
        if cfg.attn_layer == "se":
            self.se = self.attn_layer(rd_ratio=cfg.se_ratio, name="se")
        else:
            self.se = self.attn_layer(name="se")
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.act2 = self.act_layer()

    def call(self, x, training=False):
        shortcut = x

        x = self.pad1(x)
        x = self.conv1(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        if self.aa is not None:
            x = self.aa(x, training=training)

        x = self.pad2(x)
        x = self.conv2(x)
        x = self.bn2(x, training=training)

        x = self.se(x, training=training)

        x = self.drop_path(x, training=training)

        if self.downsample_layer is not None:
            shortcut = self.downsample_layer(shortcut, training=training)
        x += shortcut
        x = self.act2(x)

        return x


class Bottleneck(tf.keras.layers.Layer):
    expansion = 4

    def __init__(
        self,
        cfg: ResNetConfig,
        nb_channels: int,
        stride: int,
        drop_path_rate: float,
        downsample_layer,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.cfg = cfg
        self.downsample_layer = downsample_layer
        self.act_layer = act_layer_factory(cfg.act_layer)
        self.norm_layer = norm_layer_factory(cfg.norm_layer)
        self.attn_layer = attn_layer_factory(cfg.attn_layer)

        # Number of channels after second convolution
        width = int(math.floor(nb_channels * (cfg.base_width / 64)) * cfg.cardinality)
        # Number of channels after first convolution
        first_planes = width // cfg.block_reduce_first
        # Number of channels after third convolution
        out_planes = nb_channels * self.expansion
        use_aa = cfg.aa_layer and stride == 2

        self.conv1 = tf.keras.layers.Conv2D(
            filters=first_planes,
            kernel_size=1,
            use_bias=False,
            name="conv1",
        )
        self.bn1 = self.norm_layer(name="bn1")
        self.act1 = self.act_layer()

        self.pad2 = tf.keras.layers.ZeroPadding2D(padding=1)
        self.conv2 = tf.keras.layers.Conv2D(
            filters=width,
            kernel_size=3,
            # If we use anti-aliasing, the anti-aliasing layer takes care of strides
            strides=1 if use_aa else stride,
            groups=cfg.cardinality,
            use_bias=False,
            name="conv2",
        )
        self.bn2 = self.norm_layer(name="bn2")
        self.act2 = self.act_layer()
        self.aa = BlurPool2D(stride=stride) if use_aa else None

        self.conv3 = tf.keras.layers.Conv2D(
            filters=out_planes,
            kernel_size=1,
            use_bias=False,
            name="conv3",
        )
        initializer = "zeros" if cfg.zero_init_last_bn else "ones"
        if cfg.norm_layer == "batch_norm":
            # Only batch norm layer has moving_variance_initializer parameter
            self.bn3 = self.norm_layer(
                gamma_initializer=initializer,
                moving_variance_initializer=initializer,
                name="bn3",
            )
        else:
            self.bn3 = self.norm_layer(gamma_initializer=initializer, name="bn3")
        if cfg.attn_layer == "se":
            self.se = self.attn_layer(rd_ratio=cfg.se_ratio, name="se")
        else:
            self.se = self.attn_layer(name="se")
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.act3 = self.act_layer()

    def call(self, x, training=False):
        shortcut = x

        x = self.conv1(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)

        x = self.pad2(x)
        x = self.conv2(x)
        x = self.bn2(x, training=training)
        x = self.act2(x)
        if self.aa is not None:
            x = self.aa(x, training=training)

        x = self.conv3(x)
        x = self.bn3(x, training=training)

        x = self.se(x, training=training)

        x = self.drop_path(x, training=training)

        if self.downsample_layer is not None:
            shortcut = self.downsample_layer(shortcut, training=training)
        x += shortcut
        x = self.act3(x)

        return x


def downsample_avg(cfg: ResNetConfig, out_channels: int, stride: int, name: str):
    norm_layer = norm_layer_factory(cfg.norm_layer)

    if stride != 1:
        pool = tf.keras.layers.AveragePooling2D(
            pool_size=2, strides=stride, padding="same"
        )
    else:
        pool = tf.keras.layers.Activation("linear")
    conv = tf.keras.layers.Conv2D(
        filters=out_channels,
        kernel_size=1,
        strides=1,
        use_bias=False,
        name=name + "/downsample/1",
    )
    bn = norm_layer(name=name + "/downsample/2")
    return tf.keras.Sequential([pool, conv, bn])


def downsample_conv(cfg: ResNetConfig, out_channels: int, stride: int, name: str):
    norm_layer = norm_layer_factory(cfg.norm_layer)

    # This layer is part of the conv layer in pytorch and so is not being tracked here
    p = (stride + cfg.down_kernel_size) // 2 - 1
    pad = tf.keras.layers.ZeroPadding2D(padding=p)

    conv = tf.keras.layers.Conv2D(
        filters=out_channels,
        kernel_size=cfg.down_kernel_size,
        strides=stride,
        use_bias=False,
        name=name + "/downsample/0",
    )
    bn = norm_layer(name=name + "/downsample/1")
    return tf.keras.Sequential([pad, conv, bn])


def make_stage(
    idx: int,
    in_channels: int,
    cfg: ResNetConfig,
    name: str,
):
    stage_name = f"layer{idx + 1}"  # Weight compatibility requires this name

    assert cfg.block in {"basic_block", "bottleneck"}
    block_cls = BasicBlock if cfg.block == "basic_block" else Bottleneck
    nb_blocks = cfg.nb_blocks[idx]
    nb_channels = cfg.nb_channels[idx]
    # The actual number of channels after the block. Not the same as nb_channels,
    # because Bottleneck blocks have an expansion factor = 4.
    out_channels = nb_channels * block_cls.expansion

    assert cfg.downsample_mode in {"avg", "conv"}
    downsample_fn = downsample_avg if cfg.downsample_mode == "avg" else downsample_conv

    # We need to know the absolute number of blocks to set stochastic depth decay
    total_nb_blocks = sum(cfg.nb_blocks)
    total_block_idx = sum(cfg.nb_blocks[:idx])

    blocks = []
    for block_idx in range(nb_blocks):
        stride = 1 if idx == 0 or block_idx > 0 else 2
        if (block_idx == 0) and (stride != 1 or in_channels != out_channels):
            downsample_layer = downsample_fn(
                cfg, out_channels, stride, name=f"{name}/{stage_name}/0"
            )
        else:
            downsample_layer = None

        # Stochastic depth linear decay rule
        block_dpr = cfg.drop_path_rate * total_block_idx / (total_nb_blocks - 1)

        blocks.append(
            block_cls(
                cfg,
                nb_channels=nb_channels,
                stride=stride,
                downsample_layer=downsample_layer,
                drop_path_rate=block_dpr,
                name=f"{stage_name}/{block_idx}",
            )
        )

        in_channels = nb_channels
        total_block_idx += 1
    return blocks, in_channels


@keras_serializable
class ResNet(tf.keras.Model):
    """
    ResNet / ResNeXt / SE-ResNeXt / SE-Net

    This class implements various ResNet versions.

    Parameters
    ----------
    nb_classes : int, default 1000
        Number of classification classes.
    in_channels : int, default 3
        Number of input (color) channels.
    input_size: Tuple[int, int], default (224, 224)
        Input image size
    block : str
        Which residual block to use. One of {"basic_block", "bottleneck"}
    nb_blocks : Tuple of int
        Numbers of blocks in each stage
    nb_channels: Tuple of int, default (64, 128, 256, 512)
        Number of channels in each stage
    cardinality : int, default 1
        Number of convolution groups for 3x3 conv in Bottleneck.
    base_width : int, default 64
        Factor determining bottleneck channels, via formula
            `channels * base_width / 64 * cardinality`
    downsample_mode: str, default "conv"
        How to downsample in the skip connection between stages. One of {"conv", "avg"}
    zero_init_last_bn: bool, default True
        Should last BN layer in each residual block be zero-initialized
    stem_width : int, default 64
        Number of channels in stem convolutions
    stem_type : str, default ""
        The type of stem:
          - "", default - a single 7x7 conv with a width of stem_width
          - "deep" - three 3x3 convolution layers of widths stem_width, stem_width,
              stem_width * 2
          - "deep_tiered" - three 3x3 conv layers of widths stem_width // 4 * 3,
              stem_width, stem_width * 2
    replace_stem_pool: bool, default False
        Replace maxpooling in stem by conv+bn+act layers
    block_reduce_first: int, default 1
        Reduction factor for first convolution output width of residual blocks,
        1 for all archs at the moment (it is to for a SE-Net in timm that is not
        implemented here).
    down_kernel_size: int, default 1
        Kernel size of residual block downsampling path, 1x1 for all archs (it is 3x3
        for a SE-Net in timm that is not implemented here)
    act_layer : str, default "relu"
        Activation layer, see `act_layer_factory` for allowed values
    norm_layer : str, default "batch_norm"
        Normalization layer, see `norm_layer_factory` for allowed values
    aa_layer : str, default ""
        Anti-aliasing layer. Can be "" (no AA-layer) or "blur_pool".
    attn_layer: str, default ""
        Attention layer, see `attn_layer_factory` for allowed values.
    se_ratio: float, default 1. / 16 = 0.0625
        Reduction ratio in SE blocks. Ignored, if attn_layer != "se".
    drop_rate : float, default 0.
        Dropout probability before classifier, for training.
    drop_path_rate: float, default 0.
        Dropout probability for DropPath layers, for training.
    global_pool : str, default "avg"
        Global pooling type. One of "avg", "max"
    """

    cfg_class = ResNetConfig

    # The blur_kernel parameter in BlurPool2D layers is non-trainable and the value
    # is set during layer construction. The pycharm weights are not stored, hence we
    # cannot load them
    keys_to_ignore_on_load_missing = ["blur_kernel"]

    def __init__(self, cfg: ResNetConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.cfg = cfg

        self.act_layer = act_layer_factory(cfg.act_layer)
        self.norm_layer = norm_layer_factory(cfg.norm_layer)

        if cfg.stem_type in {"deep", "deep_tiered"}:
            in_channels = cfg.stem_width * 2
            if cfg.stem_type == "deep_tiered":
                stem_chns = (3 * (cfg.stem_width // 4), cfg.stem_width)
            else:
                stem_chns = (cfg.stem_width, cfg.stem_width)
            self.pad1 = tf.keras.layers.ZeroPadding2D(padding=1)
            conv1_0 = tf.keras.layers.Conv2D(
                filters=stem_chns[0],
                kernel_size=3,
                strides=2,
                use_bias=False,
                name=f"{self.name}/conv1/0",
            )
            bn1_0 = self.norm_layer(name=f"{self.name}/conv1/1")
            act1_0 = self.act_layer()
            conv1_1 = tf.keras.layers.Conv2D(
                filters=stem_chns[1],
                kernel_size=3,
                padding="same",
                use_bias=False,
                name=f"{self.name}/conv1/3",
            )
            bn1_1 = self.norm_layer(name=f"{self.name}/conv1/4")
            act1_1 = self.act_layer()
            conv1_2 = tf.keras.layers.Conv2D(
                filters=in_channels,
                kernel_size=3,
                padding="same",
                use_bias=False,
                name=f"{self.name}/conv1/6",
            )
            self.conv1 = tf.keras.Sequential(
                [conv1_0, bn1_0, act1_0, conv1_1, bn1_1, act1_1, conv1_2]
            )
        else:
            in_channels = 64
            # In TF "same" padding with strides != 1 is not the same as (3, 3) padding
            # in pytorch, hence the need for an explicit padding layer
            self.pad1 = tf.keras.layers.ZeroPadding2D(padding=3)
            self.conv1 = tf.keras.layers.Conv2D(
                filters=in_channels,
                kernel_size=7,
                strides=2,
                use_bias=False,
                name="conv1",
            )
        self.bn1 = self.norm_layer(name="bn1")
        self.act1 = self.act_layer()

        # Stem Pooling
        if cfg.replace_stem_pool:
            # Note that if replace_stem_pool=True, we are ignoring the aa_layer
            # None of the timm models use both.
            pad = tf.keras.layers.ZeroPadding2D(padding=1)
            conv = tf.keras.layers.Conv2D(
                filters=in_channels,
                kernel_size=3,
                strides=2,
                use_bias=False,
                name=f"{self.name}/maxpool/0",
            )
            bn = self.norm_layer(name=f"{self.name}/maxpool/1")
            act = self.act_layer()
            self.maxpool = tf.keras.Sequential([pad, conv, bn, act])
        else:
            if cfg.aa_layer:
                pad = tf.keras.layers.ZeroPadding2D(padding=1)
                pool = tf.keras.layers.MaxPool2D(pool_size=3, strides=1)
                aa = BlurPool2D(stride=2, name=f"{self.name}/maxpool/2")
                self.maxpool = tf.keras.Sequential([pad, pool, aa])
            else:
                pad = tf.keras.layers.ZeroPadding2D(padding=1)
                pool = tf.keras.layers.MaxPool2D(pool_size=3, strides=2)
                self.maxpool = tf.keras.Sequential([pad, pool])

        self.blocks = []
        for idx in range(4):
            stage_blocks, in_channels = make_stage(
                idx=idx, in_channels=in_channels, cfg=cfg, name=self.name
            )
            self.blocks.extend(stage_blocks)

        # Head (pooling and classifier)
        self.head = ClassifierHead(
            nb_classes=cfg.nb_classes,
            pool_type=cfg.global_pool,
            drop_rate=cfg.drop_rate,
            use_conv=False,
            name="remove",
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        return (
            ["stem"]
            + [f"block_{j}" for j in range(sum(self.cfg.nb_blocks))]
            + ["features", "logits"]
        )

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        x = self.pad1(x)
        x = self.conv1(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        x = self.maxpool(x)
        features["stem"] = x

        for j, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{j}"] = x
        features["features"] = x

        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x, training=training)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def resnet18():
    """Constructs a ResNet-18 model."""
    cfg = ResNetConfig(
        name="resnet18", url="[timm]", block="basic_block", nb_blocks=(2, 2, 2, 2)
    )
    return ResNet, cfg


@register_model
def resnet18d():
    """Constructs a ResNet-18-D model."""
    cfg = ResNetConfig(
        name="resnet18d",
        url="[timm]",
        block="basic_block",
        nb_blocks=(2, 2, 2, 2),
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet26():
    """Constructs a ResNet-26 model."""
    cfg = ResNetConfig(
        name="resnet26",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(2, 2, 2, 2),
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def resnet26d():
    """Constructs a ResNet-26-D model."""
    cfg = ResNetConfig(
        name="resnet26d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(2, 2, 2, 2),
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet26t():
    """Constructs a ResNet-26-T model."""
    cfg = ResNetConfig(
        name="resnet26t",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(2, 2, 2, 2),
        pool_size=8,
        stem_width=32,
        stem_type="deep_tiered",
        downsample_mode="avg",
        crop_pct=0.94,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet34():
    """Constructs a ResNet-34 model."""
    cfg = ResNetConfig(
        name="resnet34", url="[timm]", block="basic_block", nb_blocks=(3, 4, 6, 3)
    )
    return ResNet, cfg


@register_model
def resnet34d():
    """Constructs a ResNet-34-D model."""
    cfg = ResNetConfig(
        name="resnet34d",
        url="[timm]",
        block="basic_block",
        nb_blocks=(3, 4, 6, 3),
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet50():
    """Constructs a ResNet-50 model."""
    cfg = ResNetConfig(
        name="resnet50",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        interpolation="bicubic",
        crop_pct=0.95,
    )
    return ResNet, cfg


@register_model
def resnet50d():
    """Constructs a ResNet-50-D model."""
    cfg = ResNetConfig(
        name="resnet50d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet101():
    """Constructs a ResNet-101 model."""
    cfg = ResNetConfig(
        name="resnet101",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        interpolation="bicubic",
        crop_pct=0.95,
    )
    return ResNet, cfg


@register_model
def resnet101d():
    """Constructs a ResNet-101-D model."""
    cfg = ResNetConfig(
        name="resnet101d",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        pool_size=8,
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        test_input_size=(320, 320),
        interpolation="bicubic",
        crop_pct=1.0,
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet152():
    """Constructs a ResNet-152 model."""
    cfg = ResNetConfig(
        name="resnet152",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 8, 36, 3),
        interpolation="bicubic",
        crop_pct=0.95,
    )
    return ResNet, cfg


@register_model
def resnet152d():
    """Constructs a ResNet-152-D model."""
    cfg = ResNetConfig(
        name="resnet152d",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(3, 8, 36, 3),
        pool_size=8,
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        test_input_size=(320, 320),
        interpolation="bicubic",
        crop_pct=1.0,
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnet200d():
    """Constructs a ResNet-200-D model."""
    cfg = ResNetConfig(
        name="resnet200d",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(3, 24, 36, 3),
        pool_size=8,
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        test_input_size=(320, 320),
        interpolation="bicubic",
        crop_pct=1.0,
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def tv_resnet34():
    """Constructs a ResNet-34 model with original Torchvision weights."""
    cfg = ResNetConfig(
        name="tv_resnet34", url="[timm]", block="basic_block", nb_blocks=(3, 4, 6, 3)
    )
    return ResNet, cfg


@register_model
def tv_resnet50():
    """Constructs a ResNet-50 model with original Torchvision weights."""
    cfg = ResNetConfig(
        name="tv_resnet50", url="[timm]", block="bottleneck", nb_blocks=(3, 4, 6, 3)
    )
    return ResNet, cfg


@register_model
def tv_resnet101():
    """Constructs a ResNet-101 model w/ Torchvision pretrained weights."""
    cfg = ResNetConfig(
        name="tv_resnet101", url="[timm]", block="bottleneck", nb_blocks=(3, 4, 23, 3)
    )
    return ResNet, cfg


@register_model
def tv_resnet152():
    """Constructs a ResNet-152 model w/ Torchvision pretrained weights."""
    cfg = ResNetConfig(
        name="tv_resnet152", url="[timm]", block="bottleneck", nb_blocks=(3, 8, 36, 3)
    )
    return ResNet, cfg


@register_model
def wide_resnet50_2():
    """Constructs a Wide ResNet-50-2 model.
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.
    """
    cfg = ResNetConfig(
        name="wide_resnet50_2",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        base_width=128,
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def wide_resnet101_2():
    """Constructs a Wide ResNet-101-2 model.
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same.
    """
    cfg = ResNetConfig(
        name="wide_resnet101_2",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        base_width=128,
    )
    return ResNet, cfg


@register_model
def resnet50_gn():
    """Constructs a ResNet-50 model w/ GroupNorm"""
    cfg = ResNetConfig(
        name="resnet50_gn",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        norm_layer="group_norm",
        crop_pct=0.94,
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def resnext50_32x4d():
    """Constructs a ResNeXt50-32x4d model."""
    cfg = ResNetConfig(
        name="resnext50_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        cardinality=32,
        base_width=4,
        crop_pct=0.95,
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def resnext50d_32x4d():
    """
    Constructs a ResNeXt50d-32x4d model. ResNext50 w/ deep stem & avg pool downsample
    """
    cfg = ResNetConfig(
        name="resnext50d_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        cardinality=32,
        base_width=4,
        stem_width=32,
        stem_type="deep",
        downsample_mode="avg",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnext101_32x8d():
    """Constructs a ResNeXt-101 32x8d model."""
    cfg = ResNetConfig(
        name="resnext101_32x8d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=8,
    )
    return ResNet, cfg


@register_model
def tv_resnext50_32x4d():
    """Constructs a ResNeXt50-32x4d model with original Torchvision weights."""
    cfg = ResNetConfig(
        name="tv_resnext50_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        cardinality=32,
        base_width=4,
    )
    return ResNet, cfg


@register_model
def ig_resnext101_32x8d():
    """Constructs a ResNeXt-101 32x8 model pretrained on weakly-supervised data
    and finetuned on ImageNet from Figure 5 in
    `"Exploring the Limits of Weakly Supervised Pretraining"
    <https://arxiv.org/abs/1805.00932>`_
    Weights from https://pytorch.org/hub/facebookresearch_WSL-Images_resnext/
    """
    cfg = ResNetConfig(
        name="ig_resnext101_32x8d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=8,
    )
    return ResNet, cfg


@register_model
def ig_resnext101_32x16d():
    """Constructs a ResNeXt-101 32x16 model pretrained on weakly-supervised data
    and finetuned on ImageNet from Figure 5 in
    `"Exploring the Limits of Weakly Supervised Pretraining"
    <https://arxiv.org/abs/1805.00932>`_
    Weights from https://pytorch.org/hub/facebookresearch_WSL-Images_resnext/
    """
    cfg = ResNetConfig(
        name="ig_resnext101_32x16d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=16,
    )
    return ResNet, cfg


@register_model
def ig_resnext101_32x32d():
    """Constructs a ResNeXt-101 32x32 model pretrained on weakly-supervised data
    and finetuned on ImageNet from Figure 5 in
    `"Exploring the Limits of Weakly Supervised Pretraining"
    <https://arxiv.org/abs/1805.00932>`_
    Weights from https://pytorch.org/hub/facebookresearch_WSL-Images_resnext/
    """
    cfg = ResNetConfig(
        name="ig_resnext101_32x32d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=32,
    )
    return ResNet, cfg


@register_model
def ig_resnext101_32x48d():
    """Constructs a ResNeXt-101 32x48 model pretrained on weakly-supervised data
    and finetuned on ImageNet from Figure 5 in
    `"Exploring the Limits of Weakly Supervised Pretraining"
    <https://arxiv.org/abs/1805.00932>`_
    Weights from https://pytorch.org/hub/facebookresearch_WSL-Images_resnext/
    """
    cfg = ResNetConfig(
        name="ig_resnext101_32x48d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=48,
    )
    return ResNet, cfg


@register_model
def ssl_resnet18():
    """Constructs a semi-supervised ResNet-18 model pre-trained on YFCC100M dataset and
    finetuned on ImageNet
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="ssl_resnet18",
        url="[timm]",
        block="basic_block",
        nb_blocks=(2, 2, 2, 2),
    )
    return ResNet, cfg


@register_model
def ssl_resnet50():
    """Constructs a semi-supervised ResNet-50 model pre-trained on YFCC100M dataset and
    finetuned on ImageNet
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="ssl_resnet50",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
    )
    return ResNet, cfg


@register_model
def ssl_resnext50_32x4d():
    """
    Constructs a semi-supervised ResNeXt-50 32x4 model pre-trained on YFCC100M dataset
    and finetuned on ImageNet
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="ssl_resnext50_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        cardinality=32,
        base_width=4,
    )
    return ResNet, cfg


@register_model
def ssl_resnext101_32x4d():
    """
    Constructs a semi-supervised ResNeXt-101 32x4 model pre-trained on YFCC100M dataset
    and finetuned on ImageNet
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="ssl_resnext101_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=4,
    )
    return ResNet, cfg


@register_model
def ssl_resnext101_32x8d():
    """
    Constructs a semi-supervised ResNeXt-101 32x8 model pre-trained on YFCC100M dataset
    and finetuned on ImageNet
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="ssl_resnext101_32x8d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=8,
    )
    return ResNet, cfg


@register_model
def ssl_resnext101_32x16d():
    """
    Constructs a semi-supervised ResNeXt-101 32x16 model pre-trained on YFCC100M dataset
    and finetuned on ImageNet
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="ssl_resnext101_32x16d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=16,
    )
    return ResNet, cfg


@register_model
def swsl_resnet18():
    """
    Constructs a semi-weakly supervised Resnet-18 model pre-trained on 1B weakly
    supervised image dataset and finetuned on ImageNet.
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="swsl_resnet18",
        url="[timm]",
        block="basic_block",
        nb_blocks=(2, 2, 2, 2),
    )
    return ResNet, cfg


@register_model
def swsl_resnet50():
    """
    Constructs a semi-weakly supervised ResNet-50 model pre-trained on 1B weakly
    supervised image dataset and finetuned on ImageNet.
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="swsl_resnet50",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
    )
    return ResNet, cfg


@register_model
def swsl_resnext50_32x4d():
    """
    Constructs a semi-weakly supervised ResNeXt-50 32x4 model pre-trained on 1B weakly
    supervised image dataset and finetuned on ImageNet.
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="swsl_resnext50_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        cardinality=32,
        base_width=4,
    )
    return ResNet, cfg


@register_model
def swsl_resnext101_32x4d():
    """
    Constructs a semi-weakly supervised ResNeXt-101 32x4 model pre-trained on 1B weakly
    supervised image dataset and finetuned on ImageNet.
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="swsl_resnext101_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=4,
    )
    return ResNet, cfg


@register_model
def swsl_resnext101_32x8d():
    """
    Constructs a semi-weakly supervised ResNeXt-101 32x8 model pre-trained on 1B weakly
    supervised image dataset and finetuned on ImageNet.
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="swsl_resnext101_32x8d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=8,
    )
    return ResNet, cfg


@register_model
def swsl_resnext101_32x16d():
    """
    Constructs a semi-weakly supervised ResNeXt-101 32x16 model pre-trained on 1B
    weakly supervised image dataset and finetuned on ImageNet.
    `"Billion-scale Semi-Supervised Learning for Image Classification"
    <https://arxiv.org/abs/1905.00546>`_
    Weights from https://github.com/facebookresearch/semi-supervised-ImageNet1K-models/
    """
    cfg = ResNetConfig(
        name="swsl_resnext101_32x16d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        cardinality=32,
        base_width=16,
    )
    return ResNet, cfg


@register_model
def ecaresnet26t():
    """
    Constructs an ECA-ResNeXt-26-T model.
    This is technically a 28 layer ResNet, like a 'D' bag-of-tricks model but with
    tiered 24, 32, 64 channels in the deep stem and ECA attn.
    """
    cfg = ResNetConfig(
        name="ecaresnet26t",
        url="[timm]",
        block="bottleneck",
        input_size=(256, 256),
        nb_blocks=(2, 2, 2, 2),
        stem_type="deep_tiered",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="eca",
        test_input_size=(320, 320),
        pool_size=8,
        crop_pct=0.95,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def ecaresnet50d():
    """Constructs a ResNet-50-D model with eca."""
    cfg = ResNetConfig(
        name="ecaresnet50d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        stem_type="deep",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="eca",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def ecaresnet50t():
    """
    Constructs an ECA-ResNet-50-T model.
    Like a 'D' bag-of-tricks model but with tiered 24, 32, 64 channels in the deep stem
    and ECA attn.
    """
    cfg = ResNetConfig(
        name="ecaresnet50t",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        stem_type="deep_tiered",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="eca",
        test_input_size=(320, 320),
        pool_size=8,
        crop_pct=0.95,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def ecaresnetlight():
    """Constructs a ResNet-50-D light model with eca."""
    cfg = ResNetConfig(
        name="ecaresnetlight",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(1, 1, 11, 3),
        stem_width=32,
        downsample_mode="avg",
        attn_layer="eca",
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def ecaresnet101d():
    """Constructs a ResNet-101-D model with eca."""
    cfg = ResNetConfig(
        name="ecaresnet101d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        stem_type="deep",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="eca",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def ecaresnet269d():
    """Constructs a ResNet-269-D model with ECA."""
    cfg = ResNetConfig(
        name="ecaresnet269d",
        url="[timm]",
        input_size=(320, 320),
        block="bottleneck",
        nb_blocks=(3, 30, 48, 8),
        stem_type="deep",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="eca",
        test_input_size=(352, 352),
        pool_size=10,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetblur50():
    """Constructs a ResNet-50 model with blur anti-aliasing"""
    cfg = ResNetConfig(
        name="resnetblur50",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        aa_layer="blur_pool",
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def resnetrs50():
    """Constructs a ResNet-RS-50 model.
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs50",
        url="[timm]",
        input_size=(160, 160),
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(224, 224),
        pool_size=5,
        crop_pct=0.91,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetrs101():
    """Constructs a ResNet-RS-101 model.
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs101",
        url="[timm]",
        input_size=(192, 192),
        block="bottleneck",
        nb_blocks=(3, 4, 23, 3),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(288, 288),
        pool_size=6,
        crop_pct=0.94,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetrs152():
    """Constructs a ResNet-RS-152 model.
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs152",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(3, 8, 36, 3),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(320, 320),
        pool_size=8,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetrs200():
    """Constructs a ResNet-RS-200 model.
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs200",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(3, 24, 36, 3),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(320, 320),
        pool_size=8,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetrs270():
    """Constructs a ResNet-RS-270 model.
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs270",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(4, 29, 53, 4),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(352, 352),
        pool_size=8,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetrs350():
    """Constructs a ResNet-RS-350 model.
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs350",
        url="[timm]",
        input_size=(288, 288),
        block="bottleneck",
        nb_blocks=(4, 36, 72, 4),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(384, 384),
        pool_size=9,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def resnetrs420():
    """Constructs a ResNet-RS-420 model
    Paper: Revisiting ResNets - https://arxiv.org/abs/2103.07579
    Pretrained weights from
    https://github.com/tensorflow/tpu/tree/bee9c4f6/models/official/resnet/resnet_rs
    """
    cfg = ResNetConfig(
        name="resnetrs420",
        url="[timm]",
        input_size=(320, 320),
        block="bottleneck",
        nb_blocks=(4, 44, 87, 4),
        stem_type="deep",
        stem_width=32,
        replace_stem_pool=True,
        downsample_mode="avg",
        attn_layer="se",
        se_ratio=0.25,
        test_input_size=(416, 416),
        pool_size=10,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def seresnet50():
    cfg = ResNetConfig(
        name="seresnet50",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        attn_layer="se",
        interpolation="bicubic",
    )
    return ResNet, cfg


@register_model
def seresnet152d():
    cfg = ResNetConfig(
        name="seresnet152d",
        url="[timm]",
        input_size=(256, 256),
        block="bottleneck",
        nb_blocks=(3, 8, 36, 3),
        stem_type="deep",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="se",
        test_input_size=(320, 320),
        pool_size=8,
        crop_pct=1.0,
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def seresnext26d_32x4d():
    """Constructs a SE-ResNeXt-26-D model.`
    This is technically a 28 layer ResNet, using the 'D' modifier from Gluon /
    bag-of-tricks for combination of deep stem and avg_pool in downsample.
    """
    cfg = ResNetConfig(
        name="seresnext26d_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(2, 2, 2, 2),
        cardinality=32,
        base_width=4,
        stem_type="deep",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="se",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def seresnext26t_32x4d():
    """Constructs a SE-ResNet-26-T model.
    This is technically a 28 layer ResNet, like a 'D' bag-of-tricks model but with
    tiered 24, 32, 64 channels in the deep stem.
    """
    cfg = ResNetConfig(
        name="seresnext26t_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(2, 2, 2, 2),
        cardinality=32,
        base_width=4,
        stem_type="deep_tiered",
        stem_width=32,
        downsample_mode="avg",
        attn_layer="se",
        interpolation="bicubic",
        first_conv="conv1/0",
    )
    return ResNet, cfg


@register_model
def seresnext50_32x4d():
    cfg = ResNetConfig(
        name="seresnext50_32x4d",
        url="[timm]",
        block="bottleneck",
        nb_blocks=(3, 4, 6, 3),
        cardinality=32,
        base_width=4,
        attn_layer="se",
        interpolation="bicubic",
    )
    return ResNet, cfg



from collections import OrderedDict
from typing import Tuple

import tensorflow as tf

from tfimm.layers import DropPath, PatchEmbeddings, norm_layer_factory

from .common import MLPBlock


def window_partition(
    x: tf.Tensor, window_size: int
) -> Tuple[tf.Tensor, Tuple[int, int]]:
    """
    Partition 4D tensor into non-overlapping windows with padding if needed.

    Args:
        x: Input tensor of shape (N, H, W, C).
        window_size: Window size.

    Returns:
        windows: Tensor after partition of shape
            (N * nb_windows, window_size, window_size, C).
        (Hp, Hw): Padded height and width to make spatial dimensions of x
            be multiples of window_size.
    """
    n, h, w, c = tf.unstack(tf.shape(x))
    pad_h = (window_size - h % window_size) % window_size
    pad_w = (window_size - w % window_size) % window_size

    x = tf.cond(
        (pad_h > 0) | (pad_w > 0),
        lambda: tf.pad(x, [(0, 0), (0, pad_h), (0, pad_w), (0, 0)]),
        lambda: x,
    )
    _, hp, wp, _ = tf.unstack(tf.shape(x))

    x = tf.reshape(
        x, (n, hp // window_size, window_size, wp // window_size, window_size, c)
    )
    x = tf.transpose(x, (0, 1, 3, 2, 4, 5))
    x = tf.reshape(x, (-1, window_size, window_size, c))
    return x, (hp, wp)


def window_unpartition(
    windows: tf.Tensor, pad_hw: Tuple[int, int], hw: Tuple[int, int]
) -> tf.Tensor:
    """
    Unpartitions windows into original sequences and removes padding.

    Args:
        windows: Tensor of shape (N * nb_windows, window_size, window_size, C).
        pad_hw: Padded height and width (Hp, Wp) of x.
        hw: Original height and width (H, W) of x before padding.

    Returns:
        x: Unpartitioned tensor of shape (B, H, W, C).
    """
    hp, wp = pad_hw
    h, w = hw[0], hw[1]
    window_size = tf.shape(windows)[1]
    nb_windows = (hp // window_size) * (wp // window_size)
    n = tf.shape(windows)[0] // nb_windows

    x = tf.reshape(
        windows, (n, hp // window_size, wp // window_size, window_size, window_size, -1)
    )
    x = tf.transpose(x, (0, 1, 3, 2, 4, 5))
    x = tf.reshape(x, (n, hp, wp, -1))

    x = tf.cond((hp > h) | (wp > w), lambda: x[:, :h, :w, :], lambda: x)
    return x


def get_rel_pos(
    q_size: int,
    k_size: int,
    rel_pos: tf.Tensor,
    interpolate_pos: bool,
) -> tf.Tensor:
    """
    Get relative positional embeddings according to the relative positions of
        query and key sizes.

    Args:
        q_size: Size of query q.
        k_size: Size of key k.
        rel_pos: Relative position embeddings (M, C).
        interpolate_pos: If True, we interpolate positional embeddings to fit size.

    Returns:
        Extracted positional embeddings according to relative positions.
    """
    m = tf.shape(rel_pos)[0]
    max_rel_dist = tf.cast(2 * tf.math.maximum(q_size, k_size) - 1, tf.int32)

    if interpolate_pos:
        # Interpolate positional embeddings if needed.
        rel_pos = tf.reshape(rel_pos, (1, m, -1))  # (1, M, C)
        rel_pos = tf.image.resize(
            rel_pos,
            size=(1, max_rel_dist),
            method="bilinear",
        )
        rel_pos = tf.reshape(rel_pos, (max_rel_dist, -1))  # (M', C)

    q_coords = tf.expand_dims(tf.range(q_size, dtype=tf.float32), axis=-1)
    k_coords = tf.expand_dims(tf.range(k_size, dtype=tf.float32), axis=0)
    # Scale the coords with short length if shapes for q and k are different.
    q_coords = q_coords * tf.cast(tf.math.maximum(k_size / q_size, 1.0), tf.float32)
    k_coords = k_coords * tf.cast(tf.math.maximum(q_size / k_size, 1.0), tf.float32)

    lambda_ = tf.cast(tf.math.maximum(q_size / k_size, 1.0), tf.float32)
    offset = tf.cast(k_size - 1, tf.float32) * lambda_
    relative_coords = (q_coords - k_coords) + offset
    relative_coords = tf.cast(relative_coords, tf.int32)
    return tf.gather(rel_pos, relative_coords)


def add_decomposed_rel_pos(
    attn: tf.Tensor,
    q: tf.Tensor,
    rel_pos_h: tf.Tensor,
    rel_pos_w: tf.Tensor,
    q_size: Tuple[int, int],
    k_size: Tuple[int, int],
    interpolate_pos: bool,
) -> tf.Tensor:
    """
    Calculate decomposed Relative Positional Embeddings from MVITV2 paper.

    Paper: MViTv2: Improved Multiscale Vision Transformers for Classification and
        Detection. https://arxiv.org/abs/2112.01526

    Code: https://github.com/facebookresearch/mvit

    Args:
        attn: Attention map, shape (N, H*W, H*W)
        q: Query q in the attention layer with shape (N, H*W, C).
        rel_pos_h: Relative position embeddings (Mh, C) for height axis.
        rel_pos_w: Relative position embeddings (Mw, C) for width axis.
        q_size: Spatial sequence size of query q with (q_h, q_w), usually (H, W).
        k_size: Spatial sequence size of key k with (k_h, k_w), usually (H, W).
        interpolate_pos: If True, we interpolate positional embeddings to fit size.

    Returns:
        attn: Attention map with added relative positional embeddings.
    """
    q_h, q_w = q_size
    k_h, k_w = k_size

    n, _, c = tf.unstack(tf.shape(q))  # (N, H*W, C)
    q = tf.reshape(q, (n, q_h, q_w, c))  # (N, H, W, C)

    r_h = get_rel_pos(q_h, k_h, rel_pos_h, interpolate_pos)  # (H, H, C)
    r_h = tf.einsum("nhwc,hkc->nhwk", q, r_h)  # (N, H, W, H)
    r_h = tf.expand_dims(r_h, axis=-1)  # (N, H, W, H, 1)

    r_w = get_rel_pos(q_w, k_w, rel_pos_w, interpolate_pos)  # (W, W, C)
    r_w = tf.einsum("nhwc,wkc->nhwk", q, r_w)  # (N, H, W, W)
    r_w = tf.expand_dims(r_w, axis=-2)  # (N, H, W, 1, W)

    attn = tf.reshape(attn, (n, q_h, q_w, k_h, k_w))  # (N, H, W, H, W)
    attn = attn + r_h + r_w
    attn = tf.reshape(attn, (n, q_h * q_w, k_h * k_w))  # (N, H*W, H*W)

    return attn


class RelPosAttention(tf.keras.layers.Layer):
    """Multi-head Attention block with relative position embeddings."""

    def __init__(
        self,
        fixed_input_size: bool,
        embed_dim: int,
        nb_heads: int,
        qkv_bias: bool,
        use_rel_pos: bool,
        drop_rate: float,
        attn_drop_rate: float,
        **kwargs,
    ):
        """
        Args:
            fixed_input_size: If False, we interpolate positional embeddings to fit
                input size.
            embed_dim: Number of input channels.
            nb_heads: Number of attention heads.
            qkv_bias:  If True, add a learnable bias to query, key, value.
            use_rel_pos: If True, add relative positional embeddings to attention.
        """
        super().__init__(**kwargs)
        self.fixed_input_size = fixed_input_size
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads
        self.qkv_bias = qkv_bias
        self.use_rel_pos = use_rel_pos
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate

        self.head_dim = self.embed_dim // self.nb_heads
        self.scale = self.head_dim**-0.5

        self.qkv = tf.keras.layers.Dense(embed_dim * 3, use_bias=qkv_bias, name="qkv")
        self.attn_drop = tf.keras.layers.Dropout(rate=attn_drop_rate)
        self.proj = tf.keras.layers.Dense(embed_dim, use_bias=True, name="proj")
        self.proj_drop = tf.keras.layers.Dropout(rate=drop_rate)

        self.rel_pos_h = None
        self.rel_pos_w = None

    def build(self, input_shape):
        # input_shape is (N, H, W, C)
        if self.use_rel_pos:
            # Initialize relative positional embeddings
            self.rel_pos_h = self.add_weight(
                shape=(2 * input_shape[1] - 1, self.head_dim),
                initializer="zeros",
                trainable=True,
                name="rel_pos_h",
            )
            self.rel_pos_w = self.add_weight(
                shape=(2 * input_shape[2] - 1, self.head_dim),
                initializer="zeros",
                trainable=True,
                name="rel_pos_w",
            )

    def call(self, x, training=False):
        n, h, w, _ = tf.unstack(tf.shape(x))  # (N, H, W, C)

        qkv = self.qkv(x)  # (N, H, W, 3*C)
        qkv = tf.reshape(qkv, (n, h * w, 3, self.nb_heads, -1))  # (N, H*W, 3, Hd, C/Hd)
        qkv = tf.transpose(qkv, (2, 0, 3, 1, 4))  # (3, N, Hd, H*W, C/Hd)
        qkv = tf.reshape(qkv, (3, n * self.nb_heads, h * w, -1))  # (3, N*Hd, H*W, C/Hd)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = tf.linalg.matmul(q, k, transpose_b=True)  # (N*Hd, H*W, H*W)
        attn *= self.scale

        if self.use_rel_pos:
            attn = add_decomposed_rel_pos(
                attn=attn,
                q=q,
                rel_pos_h=self.rel_pos_h,
                rel_pos_w=self.rel_pos_w,
                q_size=(h, w),
                k_size=(h, w),
                interpolate_pos=not self.fixed_input_size,
            )

        attn = tf.nn.softmax(attn, axis=-1)  # (N*Hd, H*W, H*W)
        attn = self.attn_drop(attn, training=training)
        x = tf.linalg.matmul(attn, v)  # (N*Hd, H*W, C/Hd)
        x = tf.reshape(x, (n, self.nb_heads, h, w, -1))  # (N, Hd, H, W, C/Hd)
        x = tf.transpose(x, (0, 2, 3, 1, 4))  # (N, H, W, Hd, C/Hd)
        x = tf.reshape(x, (n, h, w, -1))  # (N, H, W, C)

        x = self.proj(x)
        x = self.proj_drop(x, training=training)
        return x


class ImageEncoderBlock(tf.keras.layers.Layer):
    """
    Transformer blocks with support for window attention and residual propagation.
    """

    def __init__(
        self,
        fixed_input_size: bool,
        embed_dim: int,
        nb_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        norm_layer: str,
        act_layer: str,
        use_rel_pos: bool,
        window_size: int,
        drop_rate: float,
        attn_drop_rate: float,
        drop_path_rate: float,
        **kwargs,
    ) -> None:
        """
        Args:
            fixed_input_size: If False, we interpolate positional embeddings to fit
                input size.
            embed_dim: Number of embedding dimensions.
            nb_heads: Number of attention heads in each ViT block.
            mlp_ratio: Ratio of mlp hidden dim to embedding dim.
            qkv_bias: If True, add a learnable bias to query, key, value.
            norm_layer: Normalization layer.
            act_layer: Activation layer.
            use_rel_pos: If True, add relative positional embeddings to the attention
                map.
            window_size: Window size for window attention blocks. If it equals 0, then
                use global attention.
        """
        super().__init__(**kwargs)
        self.fixed_input_size = fixed_input_size
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.norm_layer = norm_layer
        self.act_layer = act_layer
        self.use_rel_pos = use_rel_pos
        self.window_size = window_size
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.drop_path_rate = drop_path_rate

        norm_layer = norm_layer_factory(norm_layer)

        self.norm1 = norm_layer(name="norm1")
        self.attn = RelPosAttention(
            fixed_input_size=self.fixed_input_size,
            embed_dim=self.embed_dim,
            nb_heads=self.nb_heads,
            qkv_bias=self.qkv_bias,
            use_rel_pos=self.use_rel_pos,
            drop_rate=self.drop_rate,
            attn_drop_rate=self.attn_drop_rate,
            name="attn",
        )
        self.drop_path1 = DropPath(drop_prob=self.drop_path_rate)
        self.norm2 = norm_layer(name="norm2")
        self.mlp = MLPBlock(
            hidden_dim=int(self.embed_dim * self.mlp_ratio),
            embed_dim=self.embed_dim,
            act_layer=self.act_layer,
            drop_rate=self.drop_rate,
            name="mlp",
        )
        self.drop_path2 = DropPath(drop_prob=self.drop_path_rate)

    def call(self, x, training=False):
        shortcut = x
        x = self.norm1(x, training=training)
        if self.window_size > 0:
            hw = tf.shape(x)[1:3]
            x, pad_hw = window_partition(x, self.window_size)

        x = self.attn(x, training=training)

        if self.window_size > 0:
            x = window_unpartition(x, pad_hw, hw)
        x = self.drop_path1(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp(x, training=training)
        x = self.drop_path2(x, training=training)
        x = shortcut + x

        return x


class ImageEncoder(tf.keras.Model):
    def __init__(
        self,
        input_size: Tuple[int, int] = (1024, 1024),
        fixed_input_size: bool = True,
        patch_size: int = 16,
        in_channels: int = 3,
        embed_dim: int = 768,
        nb_blocks: int = 12,
        nb_heads: int = 12,
        mlp_ratio: float = 4.0,
        out_channels: int = 256,
        qkv_bias: bool = True,
        norm_layer: str = "layer_norm",
        act_layer: str = "gelu",
        use_abs_pos: bool = True,
        use_rel_pos: bool = False,
        global_attn_indices: Tuple[int, ...] = (),
        window_size: int = 0,
        drop_rate: float = 0.0,
        attn_drop_rate: float = 0.0,
        drop_path_rate: float = 0.0,
        **kwargs,
    ):
        """
        Args:
            input_size: Input image size.
            fixed_input_size: If False, we allow arbitrary input sizes and interpolate
                the positional embeddings accordingly. If True, only one input size
                is supported.
            patch_size: Patch size.
            in_channels: Number of input image channels.
            embed_dim: Patch embedding dimension.
            nb_blocks: Depth of ViT.
            nb_heads: Number of attention heads in each ViT block.
            mlp_ratio: Ratio of mlp hidden dim to embedding dim.
            out_channels: Number of channels in image embedding.
            qkv_bias: If True, add a learnable bias to query, key, value.
            norm_layer: Normalization layer.
            act_layer: Activation layer.
            use_abs_pos: If True, use absolute positional embeddings.
            use_rel_pos: If True, add relative positional embeddings to attention.
            global_attn_indices: Indexes for blocks using global attention.
            window_size: Window size for window attention blocks.
            drop_rate: Dropout rate
            attn_drop_rate: Attention dropout rate
            drop_path_rate: Dropout rate for stochastic depth
        """
        super().__init__(**kwargs)
        self.input_size = input_size
        self.fixed_input_size = fixed_input_size
        self.patch_size = patch_size
        self.in_channels = in_channels
        self.embed_dim = embed_dim
        self.nb_blocks = nb_blocks
        self.nb_heads = nb_heads
        self.mlp_ratio = mlp_ratio
        self.out_channels = out_channels
        self.qkv_bias = qkv_bias
        self.norm_layer = norm_layer
        self.act_layer = act_layer
        self.use_abs_pos = use_abs_pos
        self.use_rel_pos = use_rel_pos
        self.global_attn_indices = global_attn_indices
        self.window_size = window_size
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.drop_path_rate = drop_path_rate

        self.patch_embed = PatchEmbeddings(
            patch_size=patch_size,
            embed_dim=embed_dim,
            flatten=False,
            name="patch_embed",
        )
        self.pos_embed = None

        self.blocks = [
            ImageEncoderBlock(
                fixed_input_size=self.fixed_input_size,
                embed_dim=self.embed_dim,
                nb_heads=self.nb_heads,
                mlp_ratio=self.mlp_ratio,
                qkv_bias=self.qkv_bias,
                drop_rate=self.drop_rate,
                attn_drop_rate=self.attn_drop_rate,
                drop_path_rate=self.drop_path_rate,
                norm_layer=self.norm_layer,
                act_layer=self.act_layer,
                use_rel_pos=self.use_rel_pos,
                window_size=(
                    self.window_size if j not in self.global_attn_indices else 0
                ),
                name=f"blocks/{j}",
            )
            for j in range(self.nb_blocks)
        ]

        # Note that this norm layer is not affected by the norm_layer parameter.
        neck_norm_layer = norm_layer_factory("layer_norm_eps_1e-6")
        self.neck = [
            tf.keras.layers.Conv2D(
                filters=self.out_channels, kernel_size=1, use_bias=False, name="neck/0"
            ),
            neck_norm_layer(name="neck/1"),
            tf.keras.layers.Conv2D(
                filters=self.out_channels,
                kernel_size=3,
                padding="same",
                use_bias=False,
                name="neck/2",
            ),
            neck_norm_layer(name="neck/3"),
        ]

    @property
    def grid_size(self):
        return (
            self.input_size[0] // self.patch_size,
            self.input_size[1] // self.patch_size,
        )

    def build(self, input_shape):
        if self.use_abs_pos:
            self.pos_embed = self.add_weight(
                shape=(1, *self.grid_size, self.embed_dim),
                initializer="zeros",
                trainable=True,
                name="pos_embed",
            )

    def call(self, x, training=False, return_features=False):
        features = OrderedDict()
        x = self.patch_embed(x, training=training)
        if self.use_abs_pos:
            if not self.fixed_input_size:
                pos_embed = tf.image.resize(
                    self.pos_embed, size=tf.shape(x)[1:3], method="bilinear"
                )
            else:
                pos_embed = self.pos_embed
            x = x + pos_embed
        features["patch_embedding"] = x

        for j, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{j}"] = x

        for j, layer in enumerate(self.neck):
            x = layer(x, training=training)
        features["neck"] = x

        return (x, features) if return_features else x



# Note that mean and std have to be multiplied by 255 before being applied to pixels.
# See PrefetchLoader class in timm/data/loader.py
IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)
IMAGENET_INCEPTION_MEAN = (0.5, 0.5, 0.5)
IMAGENET_INCEPTION_STD = (0.5, 0.5, 0.5)



from . import architectures  # noqa: F401
from .models.factory import create_model, create_preprocessing  # noqa: F401
from .models.registry import list_models  # noqa: F401
from .utils import (  # noqa: F401
    cached_model_path,
    clear_model_cache,
    get_dir,
    list_cached_models,
    set_dir,
    set_model_cache,
)
from .version import __version__  # noqa: F401



"""
We provide an implementation and pretrained weights for the Segment Anything Models.

Paper: Segment Anything.
`[arXiv:2304.02643] <https://arxiv.org/abs/2304.02643>`_.

Original pytorch code and weights from
`Facebook Research <https://github.com/facebookresearch/segment-anything>`_.

The following models are available.

* ``sam_vit_b``
* ``sam_vit_l``
* ``sam_vit_h``

In the code we are trying to follow this convention in comments and docstrings.

* N is the batch dimension
* (H0, W0) is the dimension of the input image to ``SAMPRedictor``. There are no
  constraints to this size as the image will be resized and padded to the model input
  dimensions.
* (H, W) is the model input size. This is (1024, 1024) for the pretrained models.
* (H', W') is the input size for mask prompts and the output size for predicted
  mask logits. For the pretrained models this is (256, 256). To be precise, this is
  calculated as H'=4*H'' and same for W'.
* (H'', W'') is the spatial dimension of image embeddings. For pretrained models
  this is (64, 64). This is calculated as H''=H/patch_size with patch_size=16.
* C is the number of image input channels. Usually C=3.
* M1 is the number of point prompts given to the model.
* M2 is the number of box prompts given to the model. The PyTorch code only
  supports M2={0, 1}, so the accuracy with multiple box prompts might be
  limited.
* M3 is the number of mask prompts given to the model. The PyTorch code only
  supports M3={0, 1}, so the accuracy with multiple mask prompts might be limited.
* M is the number of tokens in the sparse embeddings returned by the prompt
  embedder. The number depends on M1 and M2.
* D is the embedding dimension, which is shared by both image, sparse and dense
  prompt embeddings. For the pretrained models this is 256.
* K is the number of masks returned by the model. This number is controlled by
  the model parameter ``nb_multimask_outputs`` (set to 3 in pretrained models). And
  also by the parameter ``multimask_output`` when calling ``SAMPredictor``.
"""
from dataclasses import dataclass
from functools import partial
from typing import Optional, Tuple

import tensorflow as tf

from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

from .image_encoder import ImageEncoder
from .mask_decoder import MaskDecoder
from .prompt_encoder import PromptEncoder
from .transformer import TwoWayTransformer

# Model registry will add each entrypoint fn to this
__all__ = ["SegmentAnythingModel", "SegmentAnythingModelConfig"]


@dataclass
class SegmentAnythingModelConfig(ModelConfig):
    """
    Configuration class for SAM models.

    Parameters:
        name: Name of the model.
        url: URL for pretrained weights.
        in_channels: Number of input image channels.
        input_size: Input image size (height, width)
        fixed_input_size: If True, the model only accepts inputs of size ``input_size``.
            If False, the models accepts arbitrary input sizes by interpolating the
            positional encodings to account for the new input size.
        embed_dim: The shared embedding dimension of image and prompt embeddings.
        nb_multimask_outputs: Number of masks predicted by the model for each prompt.
        mask_threshold: Threshold for thresholding mask logits to a boolean mask.

        encoder_patch_size: Patchifying the image is implemented via a convolutional
            layer with kernel size and stride equal to ``patch_size``.
        encoder_embed_dim: Feature dimensions at each stage. These are hidden feature
            dimensions. The output dimension (which has to be compatible with the
            prompt embedding dimension) is given by ``embed_dim``.
        encoder_nb_blocks: Number of attention blocks in the image encoder.
        encoder_nb_heads: Number of self-attention heads in the image encoder.
        encoder_mlp_ratio: Ratio of mlp hidden dim to embedding dim

        encoder_drop_rate: Dropout rate
        encoder_attn_drop_rate: Attention dropout rate
        encoder_drop_path_rate: Dropout rate for stochastic depth

        encoder_norm_layer: Normalization layer. See :func:`~norm_layer_factory` for
            possible values.
        encoder_act_layer: Activation function. See :func:`~act_layer_factory` for
            possible values.
        encoder_qkv_bias: If True, add bias for qkv projection layers.
        encoder_global_attn_indices: Indexes for blocks using global attention. All
            other blocks use window attention.
        encoder_window_size: Window size for window attention blocks.

        prompt_mask_hidden_dim: Hidden dimension in the mask encoder network.

        decoder_nb_blocks: Number of attention blocks in the mask decoder.
        decoder_nb_heads: Number of self-attention heads in the mask decoder.
        decoder_mlp_channels: Number of channels in mlp layers.
        decoder_iou_head_depth: Number of layers in score predictor network.
        decoder_iou_hidden_dim: Number of hidden dimensions in score predictor network.

        mean: Defines preprocessing function. If ``x`` is an image with pixel values
            in (0, 1), the preprocessing function is ``(x - mean) / std``.
        std: Defines preprpocessing function.

        first_conv: Name of first convolutional layer. Used by
            :func:`~tfimm.create_model` to adapt the number in input channels when
            loading pretrained weights.
    """

    in_channels: int = 3
    input_size: Tuple[int, int] = (1024, 1024)
    fixed_input_size: bool = True
    embed_dim: int = 256
    nb_multimask_outputs: int = 3
    mask_threshold: float = 0.0

    # Image encoder
    encoder_patch_size: int = 16
    encoder_embed_dim: int = 768
    encoder_nb_blocks: int = 12
    encoder_nb_heads: int = 12
    encoder_mlp_ratio: float = 4.0

    encoder_drop_rate: float = 0.0
    encoder_attn_drop_rate: float = 0.0
    encoder_drop_path_rate: float = 0.0

    encoder_norm_layer: str = "layer_norm_eps_1e-6"
    encoder_act_layer: str = "gelu"
    encoder_qkv_bias: bool = True
    encoder_global_attn_indices: Tuple = (2, 5, 8, 11)
    encoder_window_size: int = 14

    # Prompt encoder
    prompt_mask_hidden_dim: int = 16

    # Mask decoder
    decoder_nb_blocks: int = 2
    decoder_nb_heads: int = 8
    decoder_mlp_channels: int = 2048
    decoder_iou_head_depth: int = 3
    decoder_iou_hidden_dim: int = 256

    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD

    # Weight transfer
    first_conv: str = "image_encoder/patch_embed/proj"

    @property
    def transform_weights(self):
        """
        Returns a dictionary of weights that need a custom transform function when
        loading to a model with a different input size. Dictionary contains weights
        and the corresponding transform function.
        """
        transforms = {"image_encoder/pos_embed": transform_pos_embed}
        # Only attention blocks that use global attention need to be transformed. For
        # the other blocks, rel_pos depends on the window size, which we assume is
        # constant
        for j in self.encoder_global_attn_indices:
            prefix = f"image_encoder/blocks/{j}/attn/rel_pos"
            transforms[prefix + "_h"] = partial(transform_rel_pos, axis=0)
            transforms[prefix + "_w"] = partial(transform_rel_pos, axis=1)
        return transforms


def transform_rel_pos(
    model, rel_pos, target_cfg: SegmentAnythingModelConfig, axis: int
):
    """
    Transform function to adapt the relative positional encodings from the attention
    blocks in the image encoder between image resolutions.
    """
    grid_dim = target_cfg.input_size[axis] // target_cfg.encoder_patch_size
    new_size = 2 * grid_dim - 1

    # rel_pos has shape (L, D), but tf.image.resize needs a 3D tensor.
    rel_pos = tf.expand_dims(rel_pos, axis=0)
    rel_pos = tf.image.resize(rel_pos, size=(1, new_size), method="bilinear")
    rel_pos = rel_pos[0]
    return rel_pos


def transform_pos_embed(model, pos_embed, target_cfg: SegmentAnythingModelConfig):
    """
    Transform function to adapt the absolute positional encodings in the image encoder
    between image resolutions.
    """
    grid_size = (
        target_cfg.input_size[0] // target_cfg.encoder_patch_size,
        target_cfg.input_size[1] // target_cfg.encoder_patch_size,
    )
    pos_embed = tf.image.resize(pos_embed, size=grid_size, method="bilinear")
    return pos_embed


@keras_serializable
class SegmentAnythingModel(tf.keras.Model):
    cfg_class = SegmentAnythingModelConfig

    def __init__(self, cfg: SegmentAnythingModelConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.cfg = cfg

        self.image_encoder = ImageEncoder(
            input_size=cfg.input_size,
            fixed_input_size=cfg.fixed_input_size,
            patch_size=cfg.encoder_patch_size,
            in_channels=cfg.in_channels,
            embed_dim=cfg.encoder_embed_dim,
            nb_blocks=cfg.encoder_nb_blocks,
            nb_heads=cfg.encoder_nb_heads,
            mlp_ratio=cfg.encoder_mlp_ratio,
            out_channels=cfg.embed_dim,
            qkv_bias=cfg.encoder_qkv_bias,
            norm_layer=cfg.encoder_norm_layer,
            act_layer=cfg.encoder_act_layer,
            use_abs_pos=True,
            use_rel_pos=True,
            window_size=cfg.encoder_window_size,
            global_attn_indices=cfg.encoder_global_attn_indices,
            drop_rate=cfg.encoder_drop_rate,
            attn_drop_rate=cfg.encoder_attn_drop_rate,
            drop_path_rate=cfg.encoder_drop_path_rate,
            name="image_encoder",
        )
        self.prompt_encoder = PromptEncoder(
            embed_dim=cfg.embed_dim,
            mask_hidden_dim=cfg.prompt_mask_hidden_dim,
            act_layer="gelu",
            name="prompt_encoder",
        )
        self.mask_decoder = MaskDecoder(
            transformer=TwoWayTransformer(
                embed_dim=cfg.embed_dim,
                nb_blocks=cfg.decoder_nb_blocks,
                nb_heads=cfg.decoder_nb_heads,
                mlp_dim=cfg.decoder_mlp_channels,
                attention_downsample_rate=2,
                act_layer="relu",
                name="transformer",
            ),
            embed_dim=cfg.embed_dim,
            nb_multimask_outputs=cfg.nb_multimask_outputs,
            iou_head_depth=cfg.decoder_iou_head_depth,
            iou_head_hidden_dim=cfg.decoder_iou_hidden_dim,
            act_layer="gelu",
            name="mask_decoder",
        )

    def grid_size(
        self, input_size: Optional[Tuple[int, int]] = None
    ) -> Tuple[int, int]:
        """
        Compute the grid size of image embeddings for the model given an image input
        size.

        Args:
            input_size: Image input size (H, W). If not provided use the input size
                from the model config.

        Returns:
            Spatial size of image embeddings (H'', W'').
        """
        input_size = input_size or self.cfg.input_size
        return (
            input_size[0] // self.cfg.encoder_patch_size,
            input_size[1] // self.cfg.encoder_patch_size,
        )

    def mask_size(
        self, input_size: Optional[Tuple[int, int]] = None
    ) -> Tuple[int, int]:
        """
        Compute the size of (low res) masks for the model given an image input size.

        Args:
            input_size: Image input size (H, W). If not provided use the input size
                from the model config.

        Returns:
            Spatial size of low resolution masks (H', W').
        """
        grid_size = self.grid_size(input_size)
        return 4 * grid_size[0], 4 * grid_size[1]

    @property
    def mask_threshold(self):
        """Threshold for thresholding logit masks to boolean masks."""
        return self.cfg.mask_threshold

    @property
    def dummy_inputs(self):
        """Returns a (nested) tensor of the correct shape for inference."""
        inputs = {
            "images": tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels)),
            "points": tf.zeros((1, 1, 2)),
            "labels": tf.zeros((1, 1)),
            "boxes": tf.zeros((1, 1, 4)),
            "masks": tf.zeros((1, 1, *self.mask_size(self.cfg.input_size))),
        }
        return inputs

    def get_image_pe(self, image_embeddings: tf.Tensor) -> tf.Tensor:
        """
        Returns image positional encodings compatible with the given image embeddings.

        Args:
            image_embeddings: Image embeddings returned by the image encoder.

        Returns:
            Image positional encoder to be passed to the mask decoder.
        """
        n, h, w, _ = tf.unstack(tf.shape(image_embeddings))
        image_pe = self.prompt_encoder.get_dense_pe((h, w))  # (H'', W'', D)
        image_pe = tf.expand_dims(image_pe, axis=0)  # (1, H'', W'', D)
        image_pe = tf.tile(image_pe, (n, 1, 1, 1))  # (N, H'', W'', D)
        return image_pe

    def postprocess_logits(self, logits, input_size, return_logits: bool):
        """
        Upscales and optionally thresholds logits returned from the mask decoder to
        segmentation masks.

        Args:
            logits: Low-resolution logits returned by the mask decoder.
            input_size: Image input size (H, W).
            return_logits: If True, we don't apply a threshold.

        Returns:
            Segmentation mask (thresholded or not) of size (H, W).
        """
        _, _, h, w = tf.unstack(tf.shape(logits))  # (N, K, H', W')
        masks = tf.transpose(logits, (0, 2, 3, 1))  # (N, H', W', K)
        masks = tf.image.resize(
            masks, size=input_size, method=tf.image.ResizeMethod.BILINEAR
        )  # (N, H, W, K)
        masks = tf.transpose(masks, (0, 3, 1, 2))  # (N, K, H, W)
        if not return_logits:
            masks = masks > self.mask_threshold
        return masks

    # TODO: Add return_features
    def call(self, inputs, training=False, multimask_output=False, return_logits=False):
        """
        Predicts masks end-to-end from provided images and prompts. If prompts are not
        known in advance, using SamPredictor is recommended over calling the model
        directly.

        Args:
            inputs: A dictionary with the following entries

             * images: An (N, H, W, C) tensor of preprocessed input images.
             * points: An (N, M1, 2) tensor of point prompts with coordinates in pixel
               space, i.e., values between 0 and H or W.
             * labels: An (N, M1) tensor of labels for point prompts. 1 indicates a
               foreground point and 0 indicates a background point.
             * boxes: An (N, M2, 4) tensor of box prompts of form (left, top, right,
               bottom) with coordinates in pixel space.
             * masks: An (N, M3, H', W') tensor of mask inputs, where M3 is
               either 1 or 0 (no mask provided).

            training: Training or inference phase?
            multimask_output: If True, we return multiple nested masks for each prompt.
            return_logits: If True, we don't threshold the upscaled mask. This is useful
                if we want to resize the mask back to original image size first and
                then apply the threshold.

        Returns:

            * Masks, an (N, K, H, W) bool tensor of binary masked predictions, where K
              is determined by the multimask_output parameter.
            * Scores, an (N, K) tensor with the model's predictions of mask quality.
            * Logits, an (N, K, H', W') tensor with low resoulution logits, where
              usually H'=H/4 and W'=W/4. This can be passed as mask input to subsequent
              iterations of prediction.
        """
        # Shape (N, H'', W'', D), where H'' = H / 16 (grid size).
        image_embeddings = self.image_encoder(inputs["images"], training=training)

        # Sparse shape: (N, M, D); Dense shape: (N, H'', W'', D).
        sparse_embeddings, dense_embeddings = self.prompt_encoder(
            inputs={
                "points": inputs["points"],
                "labels": inputs["labels"],
                "boxes": inputs["boxes"],
                "masks": inputs["masks"],
            },
            training=training,
        )

        # Logits shape: (N, K, H', W'); Scores shape: (N, K).
        logits, scores = self.mask_decoder(
            inputs={
                "image_embeddings": image_embeddings,
                "image_pe": self.get_image_pe(image_embeddings),
                "sparse_embeddings": sparse_embeddings,
                "dense_embeddings": dense_embeddings,
            },
            training=training,
            multimask_output=multimask_output,
        )

        masks = self.postprocess_logits(
            logits,
            input_size=tf.shape(inputs["images"])[1:3],
            return_logits=return_logits,
        )
        return masks, scores, logits


@register_model
def sam_vit_b():
    """SAM ViT-Base"""
    cfg = SegmentAnythingModelConfig(
        name="sam_vit_b",
        url=(
            "[pytorch]https://dl.fbaipublicfiles.com/segment_anything/"
            "sam_vit_b_01ec64.pth"
        ),
        encoder_embed_dim=768,
        encoder_nb_blocks=12,
        encoder_nb_heads=12,
        encoder_global_attn_indices=(2, 5, 8, 11),
    )
    return SegmentAnythingModel, cfg


@register_model
def sam_vit_l():
    """SAM ViT-Large"""
    cfg = SegmentAnythingModelConfig(
        name="sam_vit_l",
        url=(
            "[pytorch]https://dl.fbaipublicfiles.com/segment_anything/"
            "sam_vit_l_0b3195.pth"
        ),
        encoder_embed_dim=1024,
        encoder_nb_blocks=24,
        encoder_nb_heads=16,
        encoder_global_attn_indices=(5, 11, 17, 23),
    )
    return SegmentAnythingModel, cfg


@register_model
def sam_vit_h():
    """SAM ViT-Huge"""
    cfg = SegmentAnythingModelConfig(
        name="sam_vit_h",
        url=(
            "[pytorch]https://dl.fbaipublicfiles.com/segment_anything/"
            "sam_vit_h_4b8939.pth"
        ),
        encoder_embed_dim=1280,
        encoder_nb_blocks=32,
        encoder_nb_heads=16,
        encoder_global_attn_indices=(7, 15, 23, 31),
    )
    return SegmentAnythingModel, cfg



"""
Convolution with Weight Standardization

Paper: Micro-Batch Training with Batch-Channel Normalization and Weight Standardization
Link: https://arxiv.org/abs/1903.10520v2
Code: https://github.com/joe-siyuan-qiao/WeightStandardization
"""
from typing import Tuple, Union

import tensorflow as tf

from tfimm.utils.etc import to_2tuple


def get_padding(
    kernel_size: Union[int, tuple],
    strides: Union[int, tuple] = 1,
    dilation_rate: Union[int, tuple] = 1,
) -> Tuple[int, int]:
    """Calculate pytorch-style symmetric padding for a convolution"""
    kernel_size = to_2tuple(kernel_size)
    strides = to_2tuple(strides)
    dilation_rate = to_2tuple(dilation_rate)
    padding = (
        ((strides[0] - 1) + dilation_rate[0] * (kernel_size[0] - 1)) // 2,
        ((strides[1] - 1) + dilation_rate[1] * (kernel_size[1] - 1)) // 2,
    )
    return padding


class PadConv2D(tf.keras.layers.Conv2D):
    """
    Conv2D with support for PyTorch-style symmetric padding. Used to convert PyTorch
    models.
    """

    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1),
        padding="valid",
        data_format=None,
        dilation_rate=(1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding if padding != "symmetric" else "valid",
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.pad = (
            tf.keras.layers.ZeroPadding2D(
                padding=get_padding(kernel_size, strides, dilation_rate)
            )
            if padding == "symmetric"
            else None
        )

    def call(self, x):
        if self.pad is not None:
            x = self.pad(x)
        x = super().call(x)
        return x


class PadDepthwiseConv2D(tf.keras.layers.DepthwiseConv2D):
    """
    DepthwiseConv2D with support for PyTorch-style symmetric padding. Used to convert
    PyTorch models.
    """

    def __init__(
        self,
        kernel_size,
        strides=(1, 1),
        padding="valid",
        depth_multiplier=1,
        data_format=None,
        dilation_rate=(1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        depthwise_initializer="glorot_uniform",
        bias_initializer="zeros",
        depthwise_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        depthwise_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            kernel_size=kernel_size,
            strides=strides,
            padding=padding if padding != "symmetric" else "valid",
            depth_multiplier=depth_multiplier,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            depthwise_initializer=depthwise_initializer,
            bias_initializer=bias_initializer,
            depthwise_regularizer=depthwise_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            depthwise_constraint=depthwise_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.pad = (
            tf.keras.layers.ZeroPadding2D(
                padding=get_padding(kernel_size, strides, dilation_rate)
            )
            if padding == "symmetric"
            else None
        )

    def call(self, x):
        if self.pad is not None:
            x = self.pad(x)
        x = super().call(x)
        return x


class StdConv2D(tf.keras.layers.Conv2D):
    """
    Conv2D with Weight Standardization. Used for BiT ResNet-V2 models.

    Paper: `Micro-Batch Training with Batch-Channel Normalization and Weight
        Standardization`
    Link: https://arxiv.org/abs/1903.10520v2
    """

    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1),
        padding="valid",
        data_format=None,
        dilation_rate=(1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        eps=1e-8,
        **kwargs
    ):
        super().__init__(
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding if padding != "symmetric" else "valid",
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.eps = eps
        self.pad = (
            tf.keras.layers.ZeroPadding2D(
                padding=get_padding(kernel_size, strides, dilation_rate)
            )
            if padding == "symmetric"
            else None
        )

    def call(self, x):
        if self.pad is not None:
            x = self.pad(x)

        orig_kernel = self.kernel
        kernel_mean = tf.math.reduce_mean(self.kernel, axis=[0, 1, 2], keepdims=True)
        kernel = self.kernel - kernel_mean
        kernel_var = tf.math.reduce_variance(self.kernel, axis=[0, 1, 2], keepdims=True)
        kernel = kernel / tf.math.sqrt(kernel_var + self.eps)
        # We change the kernel temporarily to make use of `call(x)` of the super-class
        self.kernel = kernel
        x = super().call(x)
        self.kernel = orig_kernel
        return x



import logging
import shutil
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import tensorflow as tf

from ..interface import ProblemBase
from ..registry import cfg_serializable, get_class


@dataclass
class DistillationConfig:
    student: Any
    student_class: str
    teacher: Any
    teacher_class: str
    optimizer: Any
    optimizer_class: str = "OptimizerFactory"
    # If `True`, we normalize both teacher and student embeddings with respect to the
    # L2-norm, before computing the loss.
    normalize_embeddings: bool = True
    # We apply weight decay by summing the squares of all trainable variables and
    # multiplying them by `weight_decay`. We are ignoring Keras weight regularizers
    # and the automatically generated model losses.
    weight_decay: float = 0.0
    mixed_precision: bool = False


@cfg_serializable
class DistillationProblem(ProblemBase):
    cfg_class = DistillationConfig

    def __init__(self, cfg: DistillationConfig, timekeeping):
        self.cfg = cfg
        self.timekeeping = timekeeping

        # Setting global state before building model
        if cfg.mixed_precision:
            tf.keras.mixed_precision.set_global_policy("mixed_float16")

        # Building the model
        student, preprocess_student = get_class(cfg.student_class)(cfg=cfg.student)()
        teacher, preprocess_teacher = get_class(cfg.teacher_class)(cfg=cfg.teacher)()
        self.student = student
        self.preprocess_student = preprocess_student
        self.teacher = teacher
        self.preprocess_teacher = preprocess_teacher

        # Training metrics
        self.avg_l2_loss = tf.keras.metrics.Mean(dtype=tf.float32)
        self.avg_reg_loss = tf.keras.metrics.Mean(dtype=tf.float32)
        self.avg_loss = tf.keras.metrics.Mean(dtype=tf.float32)

        # Optimizer
        self.optimizer = get_class(cfg.optimizer_class)(
            cfg=cfg.optimizer,
            timekeeping=timekeeping,
            mixed_precision=cfg.mixed_precision,
        )()

    def ckpt_variables(self, model_only: bool = False):
        """Return dictionary with all variables that need to be added to checkpoint."""
        variables = {"model": self.student}
        if not model_only:
            variables["avg_l2_loss"] = self.avg_l2_loss
            variables["avg_reg_loss"] = self.avg_reg_loss
            variables["avg_loss"] = self.avg_loss
            variables["optimizer"] = self.optimizer
        return variables

    def start_epoch(self):
        """Called at the beginning of an epoch. Used to reset moving averages."""
        self.avg_l2_loss.reset_states()
        self.avg_reg_loss.reset_states()
        self.avg_loss.reset_states()

    def train_step(self, img, it):
        """Perform one step of training."""
        l2_loss, reg_loss = self.train_step_inner(img)

        self.avg_l2_loss.update_state(l2_loss)
        self.avg_reg_loss.update_state(reg_loss)
        self.avg_loss.update_state(reg_loss + l2_loss)

        logs = {
            "train/l2_loss": l2_loss,
            "train/reg_loss": reg_loss,
            "train/loss": l2_loss + reg_loss,
            "train/lr": self.optimizer.learning_rate(self.optimizer.iterations),
            "train/steps": self.optimizer.iterations,
        }
        return self.avg_loss.result().numpy(), logs

    @tf.function
    def train_step_inner(self, img):
        img_student = self.preprocess_student(img)
        img_teacher = self.preprocess_teacher(img)
        emb_teacher = self.teacher(img_teacher, training=False)
        emb_teacher = tf.cast(emb_teacher, tf.float32)
        # Axes over which to normalize embeddings
        emb_axes = tf.range(1, tf.rank(emb_teacher))
        if self.cfg.normalize_embeddings:
            emb_teacher = tf.math.l2_normalize(emb_teacher, axis=emb_axes)

        with tf.GradientTape() as tape:
            emb_student = self.student(img_student, training=True)
            emb_student = tf.cast(emb_student, tf.float32)
            if self.cfg.normalize_embeddings:
                emb_student = tf.math.l2_normalize(emb_student, axis=emb_axes)

            # First we compute the l2-norm for each embedding
            l2_loss = tf.square(emb_student - emb_teacher)
            l2_loss = tf.reduce_sum(l2_loss, axis=emb_axes)
            # And then we average over the batch dimension
            l2_loss = tf.reduce_mean(l2_loss)

            # Weight decay
            reg_loss = 0.0
            nb_variables = 0
            for weight in self.student.trainable_variables:
                reg_loss += tf.reduce_sum(tf.square(weight))
                nb_variables += tf.size(weight)
            reg_loss /= tf.cast(nb_variables, tf.float32)
            reg_loss *= self.cfg.weight_decay

            # Total loss
            loss = l2_loss + reg_loss
            if self.cfg.mixed_precision:
                loss = self.optimizer.get_scaled_loss(loss)

        grads = tape.gradient(loss, self.student.trainable_variables)
        if self.cfg.mixed_precision:
            grads = self.optimizer.get_unscaled_gradients(grads)
        self.optimizer.apply_gradients(zip(grads, self.student.trainable_variables))

        return l2_loss, reg_loss

    def validation(self, dataset):
        if hasattr(dataset, "validation"):
            # Validation is done by the dataset. We need to construct a callable model
            # and pass it to the dataset, which will do everything for us.
            def model(img):
                img = self.preprocess_student(img)
                emb = self.student(img, training=False)
                emb = tf.cast(emb, tf.float32)
                return emb

            return dataset.validation(model)
        else:
            # We don't have any default validation implemented at the moment.
            return 0.0, {}

    def save_model(self, save_dir):
        """Save models ready for inference."""
        save_dir = Path(save_dir)

        # We need to set policy to float32 for saving, otherwise we save models that
        # perform inference with float16, which is extremely slow on CPUs
        old_policy = tf.keras.mixed_precision.global_policy()
        tf.keras.mixed_precision.set_global_policy("float32")
        # After changing the policy, we need to create a new model using the policy
        model_factory = get_class(self.cfg.student_class)(cfg=self.cfg.student)
        save_model, _ = model_factory()
        save_model.set_weights(self.student.get_weights())

        model_dir = save_dir / "model"
        model_dir.mkdir(parents=True, exist_ok=True)
        with tempfile.TemporaryDirectory() as tmpdir:
            # If `save_dir` points to a network file system or S3FS, sometimes TF saving
            # can be very slow. It is faster to save to a temporary directory first and
            # copying data across.
            local_dir = Path(tmpdir) / "model"
            save_model.save(str(local_dir))
            # TODO: Add support for S3 paths here...
            try:
                shutil.copytree(str(local_dir), str(model_dir), dirs_exist_ok=True)
            except shutil.Error:
                logging.info("Unexpected error when copying model from tmpdir.")

        # Restore original float policy
        tf.keras.mixed_precision.set_global_policy(old_policy)



"""
Implementations of various drop layers. For now only DropPath (stochastic depth).

Copyright 2021 Martins Bruveris
Copyright 2020 Ross Wightman
"""
import tensorflow as tf


class DropPath(tf.keras.layers.Layer):
    """
    Per sample stochastic depth when applied in main path of residual blocks

    This is the same as the DropConnect created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in
    a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956
    Following the usage in `timm` we've changed the name to `drop_path`.
    """

    def __init__(self, drop_prob=None, **kwargs):
        super().__init__(**kwargs)
        self.drop_prob = drop_prob
        self.keep_prob = 1.0 - self.drop_prob

    def call(self, x, training=False):
        if not training or not self.drop_prob > 0.0:
            return x

        # Compute drop_connect tensor
        shape = (tf.shape(x)[0],) + (1,) * (len(tf.shape(x)) - 1)
        random_tensor = self.keep_prob + tf.random.uniform(shape, dtype=x.dtype)
        binary_tensor = tf.floor(random_tensor)

        # Rescale output to preserve batch statistics
        x = tf.math.divide(x, self.keep_prob) * binary_tensor
        return x



import math
from time import time
from typing import Optional, Tuple

import tensorflow as tf

from tfimm.models import registry
from tfimm.models.factory import create_model
from tfimm.utils import to_2tuple


def _below_resolution(
    lower: int,
    upper: int,
    resolution_abs: int,
    resolution_rel: Optional[float],
):
    """We check if (upper - lower) <= resolution."""
    if resolution_rel is not None:
        if abs(upper - lower) <= upper * resolution_rel:
            return True

    # Absolute resolution is always at least 1
    if abs(upper - lower) <= resolution_abs:
        return True

    return False


def _time_function(fun, img, nb_batches, verbose):
    """Helper function to time the execution of `fun(img)`."""
    # We ignore the first run because graph compilation takes time. And some memory.
    fun(img)

    # Now we start counting batches
    start = time()
    for j in range(nb_batches):
        fun(img)
        if verbose:
            print(f"Batch {j}: {(time() - start) / (j + 1):.3f}sec.")
    duration = time() - start
    return duration


def time_model(
    model_name,
    target,
    input_size,
    nb_classes,
    batch_size,
    float_policy,
    nb_batches,
    verbose=False,
):
    """
    Time backpropagation speed of model. The loss is simply the mean of all model
    outputs.

    Args:
        model_name: Model to be timed, will be created using `create_model`.
        target: One of "inference" or "backprop"
        input_size: Model input size
        nb_classes: Number of classes
        batch_size: Batch size to be used for testing.
        float_policy: Can be "float32" or "mixed_float16"
        nb_batches: Backpropagation time is averages over `nb_batches` calls.
        verbose: If `True`, we print duration of each batch

    Returns:
        Backpropagation throughput in img/sec.
    """
    assert float_policy in {"float32", "mixed_float16"}

    tf.keras.backend.clear_session()  # Release GPU memory
    # Need to set policy before creating model
    tf.keras.mixed_precision.set_global_policy(float_policy)
    dtype = "float32" if float_policy == "float32" else "float16"

    input_size = to_2tuple(input_size) if input_size is not None else input_size
    model = create_model(model_name, input_size=input_size, nb_classes=nb_classes)
    img = tf.ones(
        (batch_size, *model.cfg.input_size, model.cfg.in_channels),
        dtype=dtype,
    )

    if target == "inference":

        @tf.function(experimental_relax_shapes=True, jit_compile=True)
        def _fun(x):
            return model(x, training=False)

    elif target == "backprop":
        optimizer = tf.optimizers.SGD(learning_rate=0.0001)

        @tf.function(experimental_relax_shapes=True)
        def _fun(x):
            with tf.GradientTape() as tape:
                output = model(x, training=True)
                # The loss is always computed in float32 in order to not lose precision
                # Here we simulate it to make profiling more accurate
                output = tf.cast(output, "float32")
                loss = tf.reduce_mean(output)
                grads = tape.gradient(loss, model.trainable_variables)
            optimizer.apply_gradients(zip(grads, model.trainable_variables))

    else:
        raise ValueError(f"Unknown target: {target}.")

    duration = _time_function(_fun, img, nb_batches, verbose)
    img_per_sec = batch_size * nb_batches / duration
    return img_per_sec


def find_max_batch_size(
    model_name: str,
    target: str = "inference",
    input_size: Optional[int] = None,
    nb_classes: Optional[int] = None,
    float_policy: str = "float32",
    nb_batches: int = 3,
    start_batch_size: int = 256,
    resolution_abs: int = 1,
    resolution_rel: Optional[float] = 0.1,
    verbose: bool = False,
) -> Tuple[int, float]:
    """
    Searches for largest batch size that fits in memory.

    Args:
        model_name: Model to validate
        target: Can be "inference" or "backprop"
        float_policy: Can be "float32" or "mixed_float16"
        nb_batches: For how many batches to run the test
        start_batch_size: First batch size to try
        resolution_abs: We stop, if upper-lower <= resolution_abs
        resolution_rel: We stop, if (upper-lower) <= upper * resolution_rel
        verbose: If True, we print information about search progress
    Returns:
        Maximum batch size that does not lead to OOM errors.
        Inference time in img/sec with that batch size
    """
    upper_limit = None
    lower_limit = 0

    # Find hard batch size cap depending on model input size. The whole batch should
    # be <0.5 GB of memory.
    cfg = registry.model_config(model_name)
    img_size = 4 * cfg.input_size[0] * cfg.input_size[1] * cfg.in_channels
    max_memory = 5 * 10**8
    # We want max batch size to be a power of 2
    max_batch_size = 2 ** math.floor(math.log2(max_memory / img_size))

    continue_search = True
    next_batch_size = min(start_batch_size, max_batch_size)
    img_per_sec = 0.0
    while continue_search:
        batch_size = next_batch_size
        if verbose:
            print(f"Trying: {batch_size}. Range: ({lower_limit}, {upper_limit})")
        try:
            img_per_sec = time_model(
                model_name=model_name,
                target=target,
                input_size=input_size,
                nb_classes=nb_classes,
                batch_size=batch_size,
                float_policy=float_policy,
                nb_batches=nb_batches,
            )
            success = True
            lower_limit = batch_size

            if batch_size >= max_batch_size:
                continue_search = False
            elif upper_limit is None:
                next_batch_size = 2 * batch_size
                next_batch_size = min(next_batch_size, max_batch_size)
            elif _below_resolution(
                lower_limit, upper_limit, resolution_abs, resolution_rel
            ):
                continue_search = False
            else:
                next_batch_size = (upper_limit + batch_size) // 2

        except (
            tf.errors.InternalError,
            # The next one catches creating models with invalid parameters
            tf.errors.InvalidArgumentError,
            tf.errors.ResourceExhaustedError,
            tf.errors.UnknownError,
        ):
            success = False
            upper_limit = batch_size
            if _below_resolution(
                lower_limit, upper_limit, resolution_abs, resolution_rel
            ):
                continue_search = False
            else:
                next_batch_size = (batch_size + lower_limit) // 2

        finally:
            if verbose:
                print(f"Batch size {batch_size}: {'valid' if success else 'oom'}")

    return lower_limit, img_per_sec



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" PyTorch - TF 2.0 general utilities."""

import logging
import re
from enum import Enum

import numpy
import tensorflow as tf
from tensorflow.python.keras import backend as K

logger = logging.getLogger(__name__)


class TransposeType(Enum):
    """
    Possible transposition types
    """

    NO = "no"
    SIMPLE = "simple"
    CONV2D = "conv2d"


def convert_tf_weight_name_to_pt_weight_name(tf_name, tf_weight_shape=None):
    """
    Convert a TF 2.0 model variable name in a pytorch model weight name.

    Conventions for TF2.0 scopes -> PyTorch attribute names conversions:

        - '$1___$2' is replaced by $2 (can be used to duplicate or remove layers in
            TF2.0 vs PyTorch)
        - '_._' is replaced by a new level separation (can be used to convert TF2.0
            lists in PyTorch nn.ModulesList)
        - '/remove/' is replaced by '/' (can be used to remove additional intermediate
            levels in TF2.0)

    return tuple with:

        - pytorch model weight name
        - transpose: `TransposeType` member indicating whether and how TF2.0 and PyTorch weights matrices should be
          transposed with regards to each other
    """
    tf_name = tf_name.replace(":0", "")  # device ids
    tf_name = re.sub(
        r"/[^/]*___([^/]*)/", r"/\1/", tf_name
    )  # '$1___$2' is replaced by $2 (can be used to duplicate or remove layers in TF2.0 vs PyTorch)
    tf_name = tf_name.replace(
        "_._", "/"
    )  # '_._' is replaced by a level separation (can be used to convert TF2.0 lists in PyTorch nn.ModulesList)
    tf_name = tf_name.replace("/remove/", "/")
    tf_name = re.sub(r"//+", "/", tf_name)  # Remove empty levels at the end
    tf_name = tf_name.split(
        "/"
    )  # Convert from TF2.0 '/' separators to PyTorch '.' separators
    # Some weights have a single name without "/" such as final_logits_bias in BART
    if len(tf_name) > 1:
        tf_name = tf_name[1:]  # Remove level zero

    # When should we transpose the weights
    if (
        tf_name[-1] in {"kernel", "depthwise_kernel"}
        and tf_weight_shape is not None
        and tf_weight_shape.rank == 4
    ):
        # A simple heuristic to detect conv layer using weight array shape
        transpose = TransposeType.CONV2D
    elif bool(
        tf_name[-1] in ["kernel", "pointwise_kernel", "depthwise_kernel"]
        or "emb_projs" in tf_name
        or "out_projs" in tf_name
    ):
        transpose = TransposeType.SIMPLE
    else:
        transpose = TransposeType.NO

    # Convert standard TF2.0 names in PyTorch names
    if tf_name[-1] in {"kernel", "depthwise_kernel", "embeddings", "gamma"}:
        tf_name[-1] = "weight"
    if tf_name[-1] == "beta":
        tf_name[-1] = "bias"

    # BatchNorm layers
    if tf_name[-1] == "moving_mean":
        tf_name[-1] = "running_mean"
    if tf_name[-1] == "moving_variance":
        tf_name[-1] = "running_var"

    # Put name together
    tf_name = ".".join(tf_name)

    return tf_name, transpose


def load_pytorch_weights_in_tf2_model(
    tf_model, pt_state_dict, tf_inputs=None, allow_missing_keys=False
):
    """Load pytorch state_dict in a TF 2.0 model."""
    if tf_inputs is None:
        tf_inputs = tf_model.dummy_inputs

    if tf_inputs is not None:
        tf_model(tf_inputs, training=False)  # Make sure model is built

    # Adapt pt state dict. TF "beta" -> PT "bias"
    # But some models have PT weight "beta" (ResMLP affine layer)
    # To fix that we need to change PT name to "bias" first...
    # Other models have PT weights "gamma" (ConvNeXt layer scale)
    old_keys = []
    new_keys = []
    for key in pt_state_dict.keys():
        new_key = None
        if key.endswith(".beta"):
            new_key = key.replace(".beta", ".bias")
        elif key.endswith(".gamma"):
            new_key = key.replace(".gamma", ".weight")
        if new_key:
            old_keys.append(key)
            new_keys.append(new_key)
    for old_key, new_key in zip(old_keys, new_keys):
        pt_state_dict[new_key] = pt_state_dict.pop(old_key)

    symbolic_weights = tf_model.trainable_weights + tf_model.non_trainable_weights
    tf_loaded_numel = 0
    weight_value_tuples = []
    all_pytorch_weights = set(list(pt_state_dict.keys()))
    missing_keys = []
    for symbolic_weight in symbolic_weights:
        sw_name = symbolic_weight.name
        name, transpose = convert_tf_weight_name_to_pt_weight_name(
            sw_name,
            tf_weight_shape=symbolic_weight.shape,
        )

        # Find associated numpy array in pytorch model state dict
        if name not in pt_state_dict:
            if allow_missing_keys:
                missing_keys.append(name)
                continue
            keys_to_ignore = getattr(tf_model, "keys_to_ignore_on_load_missing", None)
            if keys_to_ignore is not None:
                # authorized missing keys don't have to be loaded
                if any(re.search(pat, sw_name) is not None for pat in keys_to_ignore):
                    continue
            print(name, sw_name)
            raise AttributeError(f"{name} not found in PyTorch model")

        array = pt_state_dict[name].numpy()

        if transpose is TransposeType.CONV2D:
            # Conv2D weight:
            #    PT: (num_out_channel, num_in_channel, kernel[0], kernel[1])
            # -> TF: (kernel[0], kernel[1], num_in_channel, num_out_channel)
            array = numpy.transpose(array, axes=(2, 3, 1, 0))
        elif transpose is TransposeType.SIMPLE:
            array = numpy.transpose(array)

        if len(symbolic_weight.shape) < len(array.shape):
            array = numpy.squeeze(array)
        elif len(symbolic_weight.shape) > len(array.shape):
            array = numpy.expand_dims(array, axis=0)

        if list(symbolic_weight.shape) != list(array.shape):
            try:
                array = numpy.reshape(array, symbolic_weight.shape)
            except (AssertionError, ValueError) as e:
                e.args += (name, sw_name)
                raise e

        try:
            assert list(symbolic_weight.shape) == list(array.shape)
        except AssertionError as e:
            e.args += (symbolic_weight.shape, array.shape)
            raise e

        tf_loaded_numel += array.size

        weight_value_tuples.append((symbolic_weight, array))
        all_pytorch_weights.discard(name)

    K.batch_set_value(weight_value_tuples)

    if tf_inputs is not None:
        tf_model(tf_inputs, training=False)  # Make sure restore ops are run

    logger.info(f"Loaded {tf_loaded_numel:,} parameters in the TF 2.0 model.")

    unexpected_keys = list(all_pytorch_weights)
    # PyTorch BN layers track number of batches, but TF does not, so these weights
    # will always be left over.
    unexpected_keys = [
        key for key in unexpected_keys if "num_batches_tracked" not in key
    ]

    if len(unexpected_keys) > 0:
        logger.warning(
            f"Some weights of the PyTorch model were not used when initializing the "
            f"TF 2.0 model {tf_model.__class__.__name__}: {unexpected_keys}."
        )
    else:
        logger.warning(
            f"All PyTorch model weights were used when initializing "
            f"{tf_model.__class__.__name__}."
        )
    if len(missing_keys) > 0:
        logger.warning(
            f"Some weights or buffers of the TF 2.0 model "
            f"{tf_model.__class__.__name__} were not initialized from the PyTorch "
            f"model and are newly initialized: {missing_keys}."
        )
    else:
        logger.warning(
            f"All the weights of {tf_model.__class__.__name__} were initialized from "
            f"the PyTorch model.\n"
        )


def load_timm_weights(model: tf.keras.Model, model_name: str):
    """Loads weights from timm in TF model."""
    try:
        import timm
    except ImportError:
        logging.error("To load weights from timm, timm needs to be installed.")
        raise

    if not timm.is_model_pretrained(model_name):
        raise ValueError(
            f"timm model {model_name} is not pretrained. Can't load weights from timm."
        )

    pt_model = timm.create_model(model_name, pretrained=True)
    pt_state_dict = pt_model.state_dict()
    load_pytorch_weights_in_tf2_model(model, pt_state_dict)


def load_pth_url_weights(model: tf.keras.Model, url: str):
    """Loads weights from a PyTorch .pth file specified by URL."""
    try:
        from torch.hub import load_state_dict_from_url
    except ImportError:
        logging.error("To load weights from pth weights, torch needs to be installed.")
        raise

    pt_state_dict = load_state_dict_from_url(url, progress=False, map_location="cpu")
    load_pytorch_weights_in_tf2_model(model, pt_state_dict)



""" EfficientNet, MobileNetV3, etc Builder

Assembles EfficieNet and related network feature blocks from string definitions.
Handles stride, dilation calculations, and selects feature extraction points.

Hacked together by / Copyright 2019, Ross Wightman
"""
import math
from collections import OrderedDict
from copy import deepcopy
from typing import List, Optional, Tuple, Union

from tfimm.utils import make_divisible

from .efficientnet_blocks import (
    BlockArgs,
    ConvBnAct,
    DepthwiseSeparableConv,
    EdgeResidual,
    InvertedResidual,
)

_DEBUG_BUILDER = False


def _log(msg):
    if _DEBUG_BUILDER:
        print(msg)


def round_channels(
    channels: int,
    multiplier: float = 1.0,
    divisor: int = 8,
    min_channels: Optional[int] = None,
    round_limit: float = 0.9,
):
    """Round number of filters based on depth multiplier."""
    return make_divisible(
        value=channels * multiplier,
        divisor=divisor,
        min_value=min_channels,
        round_limit=round_limit,
    )


def _scale_stage_depth(
    stack_args: List[BlockArgs],
    depth_multiplier=1.0,
    depth_trunc="ceil",
) -> List[BlockArgs]:
    """
    Per-stage depth scaling.

    Scales the block repeats in each stage. This depth scaling impl maintains
    compatibility with the EfficientNet scaling method, while allowing sensible
    scaling for other models that may have multiple block arg definitions in each stage.

    Returns a list of BlockArgs already repeated the correct number of times.
    """
    repeats = [ba.nb_repeats for ba in stack_args]

    # We scale the total repeat count for each stage, there may be multiple block arg
    # defs per stage so we need to sum.
    nb_repeats = sum(repeats)
    if depth_trunc == "round":
        # Truncating to int by rounding allows stages with few repeats to remain
        # proportionally smaller for longer. This is a good choice when stage
        # definitions include single repeat stages that we'd prefer to keep that way as
        # long as possible.
        nb_repeats_scaled = max(1, round(nb_repeats * depth_multiplier))
    else:
        # The default for EfficientNet truncates repeats to int via "ceil". Any
        # multiplier > 1.0 will result in an increased depth for every stage.
        nb_repeats_scaled = int(math.ceil(nb_repeats * depth_multiplier))

    # Proportionally distribute repeat count scaling to each block definition in the
    # stage. Allocation is done in reverse as it results in the first block being less
    # likely to be scaled. The first block makes less sense to repeat in most of the
    # arch definitions.
    repeats_scaled = []
    for r in repeats[::-1]:
        rs = max(1, round((r / nb_repeats * nb_repeats_scaled)))
        repeats_scaled.append(rs)
        nb_repeats -= r
        nb_repeats_scaled -= rs
    repeats_scaled = repeats_scaled[::-1]

    # Apply the calculated scaling to each block arg in the stage
    sa_scaled = []
    for ba, rep in zip(stack_args, repeats_scaled):
        sa_scaled.extend([deepcopy(ba) for _ in range(rep)])
    return sa_scaled


def decode_architecture(
    architecture: Tuple[Tuple[str, ...], ...],
    depth_multiplier: Union[float, Tuple[float, ...]] = 1.0,
    depth_truncation: str = "ceil",
    experts_multiplier: int = 1,
    fix_first_last: bool = False,
    group_size: Optional[int] = None,
) -> List[List[BlockArgs]]:
    """
    Decode block architecture definition strings into block kwargs.

    Args:
        architecture: Architecture definition strings, tuple of a tuple of strings
        depth_multiplier: Network depth multiplier
        depth_truncation: Networ depth truncation mode when applying multiplier
        experts_multiplier: CondConv experts multiplier
        fix_first_last: Fix first and last block depths when multiplier is applied
        group_size: Group size override for all blocks that weren't explicitly set in
            architecture string

    Returns:
        List of lists of block args
    """
    if isinstance(depth_multiplier, tuple):
        assert len(depth_multiplier) == len(architecture)
    else:
        depth_multiplier = (depth_multiplier,) * len(architecture)

    arch_args = []
    for stack_idx, (block_strings, multiplier) in enumerate(
        zip(architecture, depth_multiplier)
    ):
        assert isinstance(block_strings, tuple)
        stack_args = []
        for block_str in block_strings:
            assert isinstance(block_str, str)
            block_args = BlockArgs.decode(block_str)
            if block_args.nb_experts is not None:
                block_args.nb_experts *= experts_multiplier
            if group_size is not None:
                block_args.group_size = group_size
            stack_args.append(block_args)

        fix_depths = fix_first_last and stack_idx in {0, len(architecture) - 1}
        mod_multiplier = 1.0 if fix_depths else multiplier
        stack_args = _scale_stage_depth(stack_args, mod_multiplier, depth_truncation)

        arch_args.append(stack_args)
    return arch_args


class EfficientNetBuilder:
    """
    Build Trunk Blocks

    Adapted from timm.
    """

    def __init__(
        self,
        output_stride=32,
        channel_multiplier: float = 1.0,
        padding="",
        se_from_exp=False,  # ???
        act_layer=None,
        norm_layer=None,
        drop_path_rate=0.0,
    ):
        self.output_stride = output_stride
        self.channel_multiplier = channel_multiplier
        self.padding = padding
        # Calculate se channel reduction from expanded (mid) chs
        self.se_from_exp = se_from_exp
        self.norm_layer = norm_layer
        self.act_layer = act_layer
        self.drop_path_rate = drop_path_rate

    def _make_block(
        self,
        block_args: BlockArgs,
        stage_idx: int,
        block_idx: int,
        total_idx: int,
        nb_blocks: int,
    ):
        block_name = f"blocks.{stage_idx}.{block_idx}"
        # Stochastic depth
        drop_path_rate = self.drop_path_rate * total_idx / nb_blocks

        block_type = block_args.block_type
        block_args.filters = round_channels(block_args.filters, self.channel_multiplier)
        if block_args.force_in_channels is not None:
            block_args.force_in_channels = round_channels(
                block_args.force_in_channels, self.channel_multiplier
            )
        block_args.padding = self.padding
        block_args.norm_layer = self.norm_layer
        # block act fn overrides the model default
        block_args.act_layer = block_args.act_layer or self.act_layer
        assert block_args.act_layer is not None

        block_args.drop_path_rate = drop_path_rate
        if block_type != "cn":
            # TODO: Add parameter se_from_exp (used in Mobilenet v3) which does not
            #       adjust se_ratio.
            block_args.se_ratio /= block_args.exp_ratio

        if block_type == "ir":
            if block_args.nb_experts is not None:
                # TODO: Not implemented yet
                _log(f"  CondConvResidual {block_idx}, Args: {str(block_args)}")
                block = CondConvResidual(cfg=block_args, name=block_name)  # noqa: F821
            else:
                _log(f"  InvertedResidual {block_idx}, Args: {str(block_args)}")
                block = InvertedResidual(cfg=block_args, name=block_name)
        elif block_type in {"ds", "dsa"}:
            _log(f"  DepthwiseSeparable {block_idx}, Args: {str(block_args)}")
            block = DepthwiseSeparableConv(cfg=block_args, name=block_name)
        elif block_type == "er":
            _log(f"  EdgeResidual {block_idx}, Args: {str(block_args)}")
            block = EdgeResidual(cfg=block_args, name=block_name)
        elif block_type == "cn":
            _log(f"  ConvBnAct {block_idx}, Args: {str(block_args)}")
            block = ConvBnAct(cfg=block_args, name=block_name)  # noqa: F821
        else:
            raise ValueError(f"Unknown block type {block_type} while building model.")

        return block

    def __call__(self, architecture: List[List[BlockArgs]]) -> OrderedDict:
        """
        Build the blocks

        Args:
            architecture: A list of lists, outer list defines stages, inner list
                contains block configuration(s).

        Return:
             OrderedDict of block layers.
        """
        _log(f"Building model trunk with {len(architecture)} stages...")
        total_block_count = sum([len(x) for x in architecture])
        total_block_idx = 0
        current_stride = 2
        current_dilation = 1
        blocks = OrderedDict()

        # outer list of block_args defines the stacks
        for stack_idx, stack_args in enumerate(architecture):
            _log(f"Stack: {stack_idx}")
            assert isinstance(stack_args, list)

            # Each stack (stage of blocks) contains a list of block arguments
            for block_idx, block_args in enumerate(stack_args):
                _log(f" Block: {block_idx}")

                assert block_args.stride in {1, 2}
                # Only the first block in any stack can have a stride > 1
                if block_idx >= 1:
                    block_args.stride = 1

                next_dilation = current_dilation
                if block_args.stride > 1:
                    next_output_stride = current_stride * block_args.stride
                    if next_output_stride > self.output_stride:
                        next_dilation = current_dilation * block_args.stride
                        block_args.stride = 1
                        _log(
                            f"  Converting stride to dilation to maintain output "
                            f"stride of{self.output_stride}."
                        )
                    else:
                        current_stride = next_output_stride
                block_args.dilation_rate = current_dilation
                if next_dilation != current_dilation:
                    current_dilation = next_dilation

                block = self._make_block(
                    block_args,
                    stage_idx=stack_idx,
                    block_idx=block_idx,
                    total_idx=total_block_idx,
                    nb_blocks=total_block_count,
                )
                blocks[f"stage_{stack_idx}/block_{block_idx}"] = block

                total_block_idx += 1  # incr global block idx (across all stacks)
        return blocks
